import numpy as np
from qtpy import QtWidgets, QtCore

import logging

from ..io.catalog import Catalog
from ..utils.widgets import AbstractWidget

__all__ = [
    'SmartSlider'
]

logger = logging.getLogger(__name__)


class SmartSliderBase(QtWidgets.QSlider):
    value_changed = QtCore.Signal()

    def __init__(self, min_value=0, max_value=100, step=1, default_value=0,
                 orientation=QtCore.Qt.Orientation.Vertical, parent=None):

        super().__init__(orientation, parent)

        self._n = int((max_value - min_value) / step) + 1
        self._arr = np.linspace(min_value, max_value, self._n)

        self.step = step
        self.default_value = default_value

        self.setRange(1, self._n)
        self.setSingleStep(1)
        self.setValue(self.default_index)

        self._index = self.default_index

        self.valueChanged[int].connect(self.value_changed_action)

    @property
    def default_index(self):
        try:
            return self.index_from_value(self.default_value)
        except (TypeError, ValueError):
            return 1

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, i: int):
        self._index = i
        self.setValue(i)

    @property
    def value(self):
        return self._arr[self.index - 1]

    def index_from_value(self, value: float):
        i = self._arr.searchsorted(value, side='right')
        return i if i > 0 else 1

    def reset(self):
        self.index = self.default_index

    def value_changed_action(self, index: int):
        self.index = index
        self.value_changed.emit()


class SmartSlider(AbstractWidget):
    value_changed = QtCore.Signal(float)
    save_button_clicked = QtCore.Signal(float)

    def __init__(self, short_name: str = 'x', full_name: str | None = None, action: str | None = None,
                 visible: bool = True, catalog_name: str | None = None, link_to: str | None = None,
                 show_text_editor: bool = False, n_decimal_places: int = 6, show_save_button: bool = False,
                 parent=None, **kwargs):

        if catalog_name is not None:
            self.short_name = catalog_name
        else:
            self.short_name = short_name

        self.full_name = short_name if full_name is None else full_name
        self.action = f"change {self.full_name}" if action is None else action

        self.catalog_name = catalog_name
        self.link_to = link_to
        self.show_text_editor = show_text_editor
        self.n_decimal_places = n_decimal_places
        self.show_save_button = show_save_button

        self._slider_kwargs = kwargs

        self._slider: SmartSliderBase | None = None
        self._label: QtWidgets.QLabel | None = None
        self._editor: QtWidgets.QLineEdit | None = None
        self._save_button: QtWidgets.QPushButton | None = None

        super().__init__(parent=parent)
        self.setHidden(not visible)

        self._default_value_fallback = self._slider.default_value

    def init_ui(self):
        # create a slider
        orientation = QtCore.Qt.Orientation.Horizontal if self.show_text_editor else QtCore.Qt.Orientation.Vertical
        self._slider = SmartSliderBase(orientation=orientation, parent=self, **self._slider_kwargs)
        self._slider.setToolTip(f'Slide to {self.action}')

        # create a label
        self._label = QtWidgets.QLabel(f'{self.short_name} =', self)

        # create a text editor
        self._editor = QtWidgets.QLineEdit(self)
        self._editor.setMaximumWidth(120)

        # create a save button
        self._save_button = QtWidgets.QPushButton('Save!', self)

        self._label.setVisible(self.show_text_editor)
        self._editor.setVisible(self.show_text_editor)
        self._save_button.setVisible(self.show_save_button)

        self._slider.value_changed.connect(self.update_from_slider)
        self._editor.returnPressed.connect(self._update_from_editor)
        self._save_button.clicked.connect(self.save_value)

    def set_layout(self):
        self.setLayout(QtWidgets.QGridLayout())
        self.set_geometry(spacing=5, margins=0)

    def populate(self):
        self.layout().addWidget(self._slider, 1, 1, 1, 1)
        self.layout().addWidget(self._label, 1, 2, 1, 1)
        self.layout().addWidget(self._editor, 1, 3, 1, 1)
        self.layout().addWidget(self._save_button, 1, 4, 1, 1)

    @property
    def value(self):
        return self._slider.value

    def set_value(self, value):
        self._slider.index = self._slider.index_from_value(value)
        self._update_editor_text()

    def _update_editor_text(self):
        self._editor.setText('{value:.{num_decimal_places}f}'.format(value=self.value,
                                                                     num_decimal_places=self.n_decimal_places))

    def update_from_slider(self):
        self._update_editor_text()
        self.value_changed.emit(self.value)

    def _update_from_editor(self):
        try:
            self._slider.index = self._slider.index_from_value(float(self._editor.text()))
        except ValueError:
            logger.error(f'Invalid {self.full_name} value: {self._editor.text()}')
            self.reset()
        else:
            # the true slider value might stay the same, which would require a manual update of the text editor
            self.update_from_slider()

    @QtCore.Slot()
    def save_value(self):
        self.save_button_clicked.emit(self.value)

    def set_default_value(self, default_value: float):
        self._slider.default_value = default_value

    def set_default_value_from_catalog(self, cat_entry: Catalog | None):
        if cat_entry is None or self.catalog_name is None:
            return

        try:
            self._slider.default_value = cat_entry.get_col(self.catalog_name)
        except KeyError as e:
            logger.warning(e)
            return

    def reset(self):
        self._slider.reset()
        self.update_from_slider()  # see comment in _update_from_editor

    def clear(self):
        self._slider.default_value = self._default_value_fallback
        self._editor.setText("")



import abc
import logging

from .ViewerElement import ViewerElement, SliderItem
from .Image2D import Image2D


__all__ = [
    "ItemLinker",
    "XAxisLinker",
    "YAxisLinker",
    "SliderLinker",
    "ColorBarLinker"
]

logger = logging.getLogger(__name__)


class ItemLinker(abc.ABC):
    allowed_widget_type: type[ViewerElement] = ViewerElement

    def _validate(self, *args):
        if any(not isinstance(a, self.allowed_widget_type) for a in args):
            return False
        return True

    @abc.abstractmethod
    def _get_link_from_cfg(self, w1: ViewerElement) -> str:
        pass

    @abc.abstractmethod
    def _link(self, w1: ViewerElement, w2: ViewerElement):
        pass

    def link(self, w1: ViewerElement, w2: ViewerElement, links: dict):
        if not self._validate(w1, w2):
            return
        if not self._get_link_from_cfg(w1) == w2.title:
            return

        if w1.title == w2.title:
            logger.error(f"Cannot link a widget to itself (widget: {w1.title}, linker: {type(self).__name__})")
            return

        self._link(w1, w2)
        links[w1.title] = w2.title
        logger.debug(f"Widgets linked (link: {w1.title} --> {w2.title}, linker: {type(self).__name__})")

    @abc.abstractmethod
    def _unlink(self, w1: ViewerElement, w2: ViewerElement):
        pass

    def unlink(self, w1: ViewerElement, w2: ViewerElement, links: dict):
        if not links.get(w1.title) == w2.title:
            return

        self._unlink(w1, w2)
        links.pop(w1.title)
        logger.debug(f"Widgets unlinked (link: {w1.title} --> {w2.title}, linker: {type(self).__name__})")


class XAxisLinker(ItemLinker):
    def _get_link_from_cfg(self, w1: ViewerElement) -> str:
        return w1.cfg.x_axis.link_to

    def _link(self, w1: ViewerElement, w2: ViewerElement):
        w1.container.setXLink(w2.title)

    def _unlink(self, w1: ViewerElement, w2: ViewerElement):
        w1.container.setXLink(None)


class YAxisLinker(ItemLinker):
    def _get_link_from_cfg(self, w1: ViewerElement) -> str:
        return w1.cfg.y_axis.link_to

    def _link(self, w1: ViewerElement, w2: ViewerElement):
        w1.container.setYLink(w2.title)

    def _unlink(self, w1: ViewerElement, w2: ViewerElement):
        w1.container.setYLink(None)


class SliderLinker(ItemLinker):
    def __init__(self, slider: SliderItem):
        super().__init__()
        self._slider: SliderItem = slider

    def _get_link_from_cfg(self, w1: ViewerElement) -> str:
        return w1.sliders[self._slider].link_to

    def _link(self, w1: ViewerElement, w2: ViewerElement):
        slider1, slider2 = w1.sliders[self._slider], w2.sliders[self._slider]
        slider1.value_changed[float].connect(slider2.set_value)
        slider2.value_changed[float].connect(slider1.set_value)

    def _unlink(self, w1: ViewerElement, w2: ViewerElement):
        slider1, slider2 = w1.sliders[self._slider], w2.sliders[self._slider]
        slider1.value_changed[float].disconnect(slider2.set_value)
        slider2.value_changed[float].disconnect(slider1.set_value)


class ColorBarLinker(ItemLinker):
    allowed_widget_type = Image2D

    def _get_link_from_cfg(self, w1: Image2D) -> str:
        return w1.cfg.color_bar.link_to

    def _link(self, w1: Image2D, w2: Image2D):
        w1.cbar.sigLevelsChanged[tuple].connect(w2.set_levels)
        w2.cbar.sigLevelsChanged[tuple].connect(w1.set_levels)

    def _unlink(self, w1: Image2D, w2: Image2D):
        w1.cbar.sigLevelsChanged[tuple].disconnect(w2.set_levels)
        w2.cbar.sigLevelsChanged[tuple].disconnect(w1.set_levels)



import dacite
from dacite.exceptions import WrongTypeError, MissingValueError

from dataclasses import asdict, dataclass
from functools import wraps
import logging
import pathlib
import shutil
import yaml

logger = logging.getLogger(__name__)


@dataclass
class LocalFile:
    directory: str
    filename: str = f"{__package__.split('.')[0]}.yml"
    full_name: str = "Local file"
    auto_backup: bool = True

    @property
    def path(self) -> pathlib.Path:
        return pathlib.Path(self.directory) / self.filename

    @staticmethod
    def missing_ok(func):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.path.exists():
                return
            func(self, *args, **kwargs)
        return wrapper

    def save(self, data: dict) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)

        if self.path.exists():
            msg = f"{self.full_name} updated (path: {self.path})"
        else:
            msg = f"{self.full_name} created (path: {self.path})"

        save_yaml(self.path, data)

        logger.debug(msg)

    @missing_ok
    def backup(self) -> None:
        dst = self.path.parent / (self.filename + '.bak')
        shutil.copy(self.path, dst)
        logger.debug(f'{self.full_name} backed up (path: {dst})')

    @missing_ok
    def delete(self) -> None:
        # backup the file
        if self.auto_backup:
            self.backup()

        # delete the file
        self.path.unlink()
        logger.debug(f'{self.full_name} deleted (path: {self.path})')


@dataclass
class Params:
    def __post_init__(self):
        self._user_file: LocalFile | None = None

    @classmethod
    def _read(cls, filename: pathlib.Path):
        params_dict = read_yaml(filename)
        return dacite.from_dict(data_class=cls, data=params_dict, config=dacite.Config(strict=True))

    @classmethod
    def read_default_params(cls, filename: str):
        return cls._read(pathlib.Path(__file__).parent.parent / 'data' / 'config' / filename)

    @classmethod
    def read_user_params(cls, file: LocalFile, default: str | None = None):
        if default is None:
            params = dacite.from_dict(data_class=cls, data={})
        else:
            params = cls.read_default_params(default)

        user_params = None

        try:
            user_params = read_yaml(file.path)
        except FileNotFoundError:
            pass
        except yaml.YAMLError:
            logger.error(f'Failed to parse `{file.path}`. The file will be overwritten.')
        else:
            try:
                # TODO: patch the user config file using dictdiffer
                user_params = dacite.from_dict(data_class=cls, data=user_params, config=dacite.Config())
            except (WrongTypeError, MissingValueError):
                user_params = None
                logger.error(f'Failed to create a dataclass from `{file.path}`. The file will be overwritten.')

        if user_params is None:
            file.delete()
        else:
            params = user_params

        params._user_file = file
        params.save()

        return params

    def replace_params(self, filename: pathlib.Path):
        try:
            params = self._read(filename)
        except Exception as e:
            logger.error(str(e))
            return None

        params._user_file = self._user_file
        return params

    def save(self, file: LocalFile | None = None):
        if file is not None:
            output_file = file
        elif self._user_file is not None:
            output_file = self._user_file
        else:
            logger.error('No output file specified')
            return

        output_file.save(data=asdict(self))

    def get_user_params_filename(self) -> str | None:
        return str(self._user_file.path.resolve()) if self._user_file is not None else None


def read_yaml(filename) -> dict:
    with open(filename, "r") as yaml_file:
        return yaml.safe_load(yaml_file) or {}  # return empty dictionary if the file is empty


def filter_none_values(data):
    if isinstance(data, dict):
        return {k: filter_none_values(v) for k, v in data.items() if v is not None}
    return data


def save_yaml(filename, data):
    with open(filename, 'w') as yaml_file:
        yaml.safe_dump(data, yaml_file, sort_keys=False)



import numpy as np

import logging
import pathlib
import re


__all__ = [
    "get_ids_from_dir"
]

logger = logging.getLogger(__name__)


def _get_id_from_filename(filename, pattern: str) -> str | None:
    """Extract the object ID from a file name using regular expressions. If more than one ID is matched to the pattern,
    return the longest match.

    @param filename: the file name to parse
    @param pattern: the regular expression used to match the ID
    @return: the matched ID
    """

    try:
        matches: list[str] = re.findall(pattern, pathlib.Path(filename).name)
    except re.error:
        return None

    if not matches:
        return None

    return max(matches, key=len)


def get_ids_from_dir(directory, id_pattern: str = r'\d+', recursive: bool = False) -> np.ndarray | None:
    """Extract IDs from a directory using regular expressions.
    @param directory: the directory where the search for IDs is performed
    @param id_pattern: the regular expression used to match the IDs
    @param recursive: if True, search the directory recursively
    @return: a list of matched IDs
    """

    data_files = sorted(pathlib.Path(directory).glob('**/*' if recursive else '*'))
    ids = [_get_id_from_filename(p, id_pattern) for p in data_files]
    ids = np.array([i for i in ids if i is not None])

    try:
        ids = ids.astype(np.int64)  # convert IDs to int
    except ValueError:
        pass

    ids = np.unique(ids)  # remove duplicates

    if not ids.size:
        logger.error("No IDs retrieved from the data directory")
        return

    return ids



"""Package applying Qt compat of PyQt6, PySide6, PyQt5 and PySide2."""
from .qt_compat import QtImportError
from .qt_version import QT_VERSION

try:
    from . import QtCore, QtGui, QtSvg, QtWidgets
except ImportError:
    from .._util import get_logger as __get_logger

    __logger = __get_logger(__name__)
    __logger.warning("Failed to import QtCore, QtGui, QtSvg and QtWidgets.")



"""Icon map that overrides standard icons."""
from ..qtpy.QtWidgets import QStyle

NEW_STANDARD_ICON_MAP = {
    QStyle.StandardPixmap.SP_ArrowBack: {"id": "arrow_upward", "rotate": 270},
    QStyle.StandardPixmap.SP_ArrowDown: {"id": "arrow_upward", "rotate": 180},
    QStyle.StandardPixmap.SP_ArrowForward: {"id": "arrow_upward", "rotate": 90},
    QStyle.StandardPixmap.SP_ArrowLeft: {"id": "arrow_upward", "rotate": 270},
    QStyle.StandardPixmap.SP_ArrowRight: {"id": "arrow_upward", "rotate": 90},
    QStyle.StandardPixmap.SP_ArrowUp: {"id": "arrow_upward"},
    QStyle.StandardPixmap.SP_BrowserReload: {"id": "refresh"},
    QStyle.StandardPixmap.SP_BrowserStop: {"id": "close"},
    QStyle.StandardPixmap.SP_CommandLink: {"id": "east"},
    QStyle.StandardPixmap.SP_DialogApplyButton: {"id": "check_circle"},
    QStyle.StandardPixmap.SP_DialogCancelButton: {"id": "cancel"},
    QStyle.StandardPixmap.SP_DialogCloseButton: {"id": "close"},
    QStyle.StandardPixmap.SP_DialogDiscardButton: {"id": "delete"},
    QStyle.StandardPixmap.SP_DialogHelpButton: {"id": "help"},
    QStyle.StandardPixmap.SP_DialogNoButton: {"id": "not_interested"},
    QStyle.StandardPixmap.SP_DialogOkButton: {"id": "check"},
    QStyle.StandardPixmap.SP_DialogOpenButton: {"id": "launch"},
    QStyle.StandardPixmap.SP_DialogResetButton: {"id": "cleaning_services"},
    QStyle.StandardPixmap.SP_DialogSaveButton: {"id": "save"},
    QStyle.StandardPixmap.SP_DialogYesButton: {"id": "circle"},
    QStyle.StandardPixmap.SP_DirHomeIcon: {"id": "home"},
    QStyle.StandardPixmap.SP_DockWidgetCloseButton: {"id": "close"},
    QStyle.StandardPixmap.SP_FileDialogBack: {"id": "arrow_upward", "rotate": 270},
    QStyle.StandardPixmap.SP_FileDialogContentsView: {"id": "search"},
    QStyle.StandardPixmap.SP_FileDialogDetailedView: {"id": "list"},
    QStyle.StandardPixmap.SP_FileDialogEnd: {"id": "drive_file_move_rtl"},
    QStyle.StandardPixmap.SP_FileDialogInfoView: {"id": "info"},
    QStyle.StandardPixmap.SP_FileDialogListView: {"id": "grid_view"},
    QStyle.StandardPixmap.SP_FileDialogNewFolder: {"id": "create_new_folder"},
    QStyle.StandardPixmap.SP_FileDialogStart: {"id": "drive_file_move"},
    QStyle.StandardPixmap.SP_FileDialogToParent: {"id": "arrow_upward"},
    QStyle.StandardPixmap.SP_MediaPause: {"id": "pause"},
    QStyle.StandardPixmap.SP_MediaPlay: {"id": "play_arrow"},
    QStyle.StandardPixmap.SP_MediaSeekBackward: {"id": "fast_rewind"},
    QStyle.StandardPixmap.SP_MediaSeekForward: {"id": "fast_forward"},
    QStyle.StandardPixmap.SP_MediaSkipBackward: {"id": "skip_previous"},
    QStyle.StandardPixmap.SP_MediaSkipForward: {"id": "skip_next"},
    QStyle.StandardPixmap.SP_MediaStop: {"id": "stop"},
    QStyle.StandardPixmap.SP_MediaVolume: {"id": "volume_up"},
    QStyle.StandardPixmap.SP_MediaVolumeMuted: {"id": "volume_mute"},
    QStyle.StandardPixmap.SP_MessageBoxQuestion: {"id": "help", "os": ["Darwin", "Linux"]},
    QStyle.StandardPixmap.SP_TitleBarCloseButton: {"id": "close"},
    QStyle.StandardPixmap.SP_TitleBarContextHelpButton: {"id": "question_mark"},
    QStyle.StandardPixmap.SP_TitleBarMaxButton: {"id": "fullscreen"},
    QStyle.StandardPixmap.SP_TitleBarMinButton: {"id": "minimize"},
    QStyle.StandardPixmap.SP_TitleBarNormalButton: {"id": "flip_to_front"},
    QStyle.StandardPixmap.SP_TitleBarShadeButton: {"id": "chevron_right", "rotate": "270"},
    QStyle.StandardPixmap.SP_TitleBarUnshadeButton: {"id": "chevron_right", "rotate": "90"},
    QStyle.StandardPixmap.SP_ToolBarHorizontalExtensionButton: {"id": "double_arrow"},
    QStyle.StandardPixmap.SP_ToolBarVerticalExtensionButton: {"id": "double_arrow", "rotate": 90},
    QStyle.StandardPixmap.SP_TrashIcon: {"id": "delete", "os": ["Windows"]},
    QStyle.StandardPixmap.SP_VistaShield: {"id": "security", "os": ["Darwin", "Linux"]},
}

if hasattr(QStyle.StandardPixmap, "SP_DialogAbortButton"):
    NEW_STANDARD_ICON_MAP[QStyle.StandardPixmap.SP_DialogAbortButton] = {"id": "not_interested"}  # type: ignore  # noqa: E501

if hasattr(QStyle.StandardPixmap, "SP_DialogIgnoreButton"):
    NEW_STANDARD_ICON_MAP[QStyle.StandardPixmap.SP_DialogIgnoreButton] = {"id": "visibility_off"}  # type: ignore  # noqa: E501

if hasattr(QStyle.StandardPixmap, "SP_DialogNoToAllButton"):
    NEW_STANDARD_ICON_MAP[QStyle.StandardPixmap.SP_DialogNoToAllButton] = {"id": "close"}  # type: ignore  # noqa: E501

if hasattr(QStyle.StandardPixmap, "SP_DialogRetryButton"):
    NEW_STANDARD_ICON_MAP[QStyle.StandardPixmap.SP_DialogRetryButton] = {"id": "refresh"}  # type: ignore  # noqa: E501

if hasattr(QStyle.StandardPixmap, "SP_DialogSaveAllButton"):
    NEW_STANDARD_ICON_MAP[QStyle.StandardPixmap.SP_DialogSaveAllButton] = {"id": "save"}  # type: ignore  # noqa: E501

if hasattr(QStyle.StandardPixmap, "SP_DialogYesToAllButton"):
    NEW_STANDARD_ICON_MAP[QStyle.StandardPixmap.SP_DialogYesToAllButton] = {"id": "done_all"}  # type: ignore  # noqa: E501

if hasattr(QStyle.StandardPixmap, "SP_LineEditClearButton"):
    NEW_STANDARD_ICON_MAP[QStyle.StandardPixmap.SP_LineEditClearButton] = {"id": "close"}  # type: ignore  # noqa: E501

if hasattr(QStyle.StandardPixmap, "SP_TabCloseButton"):
    NEW_STANDARD_ICON_MAP[QStyle.StandardPixmap.SP_TabCloseButton] = {"id": "close"}  # type: ignore  # noqa: E501



from qtpy import QtGui, QtCore, QtWidgets

import logging
import pathlib

from ..config import config
from ..io.inspection_data import InspectionData
from ..utils.widgets import AbstractWidget

logger = logging.getLogger(__name__)


class ToolBar(QtWidgets.QToolBar, AbstractWidget):
    navigation_button_clicked = QtCore.Signal(str, bool)
    star_button_clicked = QtCore.Signal()
    reset_view_button_clicked = QtCore.Signal()
    reset_layout_button_clicked = QtCore.Signal()
    screenshot_button_clicked = QtCore.Signal()
    settings_button_clicked = QtCore.Signal()

    NAVIGATION_BUTTON_PARAMS = {'previous': {'shortcut': QtGui.QKeySequence.MoveToPreviousChar,
                                             'icon': 'arrow-backward'},
                                'next': {'shortcut': QtGui.QKeySequence.MoveToNextChar,
                                         'icon': 'arrow-forward'},
                                'previous starred': {'icon': 'arrow-backward-starred'},
                                'next starred': {'icon': 'arrow-forward-starred'}
                                }

    def __init__(self, appearance: config.Appearance, parent=None):
        self.appearance = appearance

        self._navigation_buttons: dict[str, QtWidgets.QAction] | None = None
        self._star_button: QtWidgets.QAction | None = None
        self._reset_view_button: QtWidgets.QAction | None = None
        self._reset_layout_button: QtWidgets.QAction | None = None
        self._screenshot_button: QtWidgets.QAction | None = None
        self._spacer: QtWidgets.QWidget | None = None
        self._settings_button: QtWidgets.QAction | None = None

        super().__init__(parent=parent)
        self.setWindowTitle('Commands Bar')

    @property
    def _viewer_connected_buttons(self):
        return tuple(self._navigation_buttons.values()) + (self._star_button, self._screenshot_button,
                                                           self._reset_view_button, self._reset_layout_button)

    def _create_navigation_buttons(self) -> dict[str, QtWidgets.QAction]:
        navig_buttons = {}
        for button_name, button_properties in self.NAVIGATION_BUTTON_PARAMS.items():
            button = QtWidgets.QAction('Go to the {} object'.format(button_name), self)

            if button_properties.get('shortcut'):
                button.setShortcut(button_properties['shortcut'])

            navig_buttons[button_name] = button

        return navig_buttons

    @staticmethod
    def _get_star_icon_name(starred=False):
        icon_name = 'star.svg' if starred else 'star-empty.svg'
        return icon_name

    def _get_icon(self, icon_name):
        icon_root_dir = pathlib.Path(__file__).parent.parent / 'data' / 'icons'
        icon_path = icon_root_dir / self.appearance.theme / icon_name
        if not icon_path.exists():
            icon_path = icon_root_dir / 'light' / icon_name

        return QtGui.QIcon(str(icon_path))

    def _set_icons(self):
        for button_name, button in self._navigation_buttons.items():
            button.setIcon(self._get_icon(self.NAVIGATION_BUTTON_PARAMS[button_name]['icon'] + '.svg'))

        self._star_button.setIcon(self._get_icon(self._get_star_icon_name()))
        self._screenshot_button.setIcon(self._get_icon('screenshot.svg'))
        self._reset_view_button.setIcon(self._get_icon('reset-view.svg'))
        self._reset_layout_button.setIcon(self._get_icon('reset-dock-state.svg'))
        self._settings_button.setIcon(self._get_icon('gear.svg'))

    def init_ui(self):
        # create buttons for switching to the next or previous object
        self._navigation_buttons = self._create_navigation_buttons()

        # create a `star` button
        self._star_button = QtWidgets.QAction(self)
        self._star_button.setShortcut('S')
        self._star_button.setToolTip('Star the object')

        # create a `screenshot` button
        self._screenshot_button = QtWidgets.QAction(self)
        self._screenshot_button.setToolTip('Take a screenshot')

        # create a `reset view` button
        self._reset_view_button = QtWidgets.QAction(self)
        self._reset_view_button.setToolTip('Reset the view')

        # create a `reset layout` button
        self._reset_layout_button = QtWidgets.QAction(self)
        self._reset_layout_button.setToolTip('Reset the layout')

        for b in self._viewer_connected_buttons:
            b.setEnabled(False)

        self._spacer = QtWidgets.QWidget(self)
        self._spacer.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        self._settings_button = QtWidgets.QAction(self)
        self._settings_button.setToolTip('GUI and Project Settings')

        self._set_icons()

        # connect button signals to slots
        for button_name, button in self._navigation_buttons.items():
            button.triggered.connect(lambda s, command=button_name.split(' ')[0], find_starred='starred' in button_name:
                                     self.navigation_button_clicked.emit(command, find_starred))

        self._star_button.triggered.connect(self.star_button_clicked.emit)
        self._reset_view_button.triggered.connect(self.reset_view_button_clicked.emit)
        self._reset_layout_button.triggered.connect(self.reset_layout_button_clicked.emit)
        self._screenshot_button.triggered.connect(self.screenshot_button_clicked.emit)
        self._settings_button.triggered.connect(self.settings_button_clicked)

    def set_layout(self):
        pass

    def populate(self):
        self.addAction(self._navigation_buttons['previous starred'])
        self.addAction(self._navigation_buttons['previous'])
        self.addAction(self._navigation_buttons['next'])
        self.addAction(self._navigation_buttons['next starred'])

        self.addSeparator()

        self.addAction(self._star_button)

        self.addSeparator()

        self.addAction(self._reset_view_button)
        self.addAction(self._reset_layout_button)
        self.addAction(self._screenshot_button)

        self.addWidget(self._spacer)

        self.addAction(self._settings_button)

    @QtCore.Slot(InspectionData)
    def load_project(self, review: InspectionData):
        for b in self._viewer_connected_buttons:
            b.setEnabled(True)

        self._navigation_buttons['previous starred'].setEnabled(review.has_data('starred'))
        self._navigation_buttons['next starred'].setEnabled(review.has_data('starred'))

    @QtCore.Slot(int, InspectionData)
    def load_object(self, j: int, review: InspectionData):
        self._star_button.setIcon(self._get_icon(self._get_star_icon_name(review.get_value(j, 'starred'))))

    @QtCore.Slot(bool, bool)
    def update_star_button_icon(self, starred: bool, has_starred: bool):
        self._star_button.setIcon(self._get_icon(self._get_star_icon_name(starred)))

        self._navigation_buttons['previous starred'].setEnabled(has_starred)
        self._navigation_buttons['next starred'].setEnabled(has_starred)



from pyqtgraph.dockarea.Container import StackedWidget
from pyqtgraph.dockarea.Dock import Dock

from abc import ABC, abstractmethod

from specvizitor.io.catalog import Catalog
from specvizitor.widgets.ViewerElement import ViewerElement


class PluginCore(ABC):
    @abstractmethod
    def override_widget_configs(self, widgets: dict[str, ViewerElement]):
        pass

    @abstractmethod
    def update_docks(self, docks: dict[str, Dock], cat_entry: Catalog | None = None):
        pass

    @abstractmethod
    def update_active_widgets(self, widgets: dict[str, ViewerElement], cat_entry: Catalog | None = None):
        pass

    @staticmethod
    def get_dock_stack(docks: dict[str, Dock]) -> StackedWidget | None:
        """Locate a stack of docks, if exists. If multiple stacked are found, return the largest stack.
        """

        dock_in_stack = None
        max_stack_count = 0

        # find the largest stack
        for i, d in enumerate(docks.values()):
            if hasattr(d.container(), 'stack'):
                stack = d.container().stack
                if stack.count() > max_stack_count:
                    dock_in_stack = d
                    max_stack_count = stack.count()

        if dock_in_stack is None:
            return None

        return dock_in_stack.container().stack

    @staticmethod
    def fix_dock_stack_labels(stack: StackedWidget):
        """
        * patching a pyqtgraph bug *
        when the dock area state is restored, the current active widget of the line map stack is changed to Line Map 1,
        however the last Line Map remains active (i.e. its label is still highlighted). therefore, when Line Map 1 is
        raised to the top, the last Line Map remains active
        """
        if stack.count() <= 1:
            return
        for i in range(stack.count()):
            stack.widget(i).label.setDim(True)
        stack.widget(0).raiseDock()



from __future__ import annotations

import platform

from ._icon.icon_engine import SvgIconEngine
from ._icon.svg import Svg
from ._resources.standard_icons import NEW_STANDARD_ICON_MAP
from .qtpy.QtGui import QIcon
from .qtpy.QtWidgets import QProxyStyle, QStyle, QStyleOption


class QDarkThemeStyle(QProxyStyle):
    """Style proxy to improve theme."""

    def __init__(self):
        """Initialize style proxy."""
        super().__init__()

    def standardIcon(  # noqa: N802
        self, standard_icon: QStyle.StandardPixmap, option: QStyleOption | None, widget
    ) -> QIcon:
        """Implement QProxyStyle.standardIcon."""
        icon_info = NEW_STANDARD_ICON_MAP.get(standard_icon)
        if icon_info is None:
            return super().standardIcon(standard_icon, option, widget)

        os_list = icon_info.get("os")
        if os_list is not None and platform.system() not in os_list:
            return super().standardIcon(standard_icon, option, widget)

        rotate = icon_info.get("rotate", 0)
        svg = Svg(icon_info["id"]).rotate(rotate)
        icon_engine = SvgIconEngine(svg)
        return QIcon(icon_engine)



from qtpy import QtCore, QtWidgets


class MyQTextEdit(QtWidgets.QTextEdit):
    def keyReleaseEvent(self, event):
        if event.key() == QtCore.Qt.Key_Escape:
            self.clearFocus()
            return
        super().keyReleaseEvent(event)



from qtpy import QtWidgets, QtCore


class Section(QtWidgets.QWidget):
    def __init__(self, title="", animation_duration=100, parent=None):
        super().__init__(parent)

        self.animationDuration = animation_duration

        self.toggleButton = QtWidgets.QToolButton(self)
        self.headerLine = QtWidgets.QFrame(self)
        self.toggleAnimation = QtCore.QParallelAnimationGroup(self)
        self.contentArea = QtWidgets.QScrollArea(self)
        self.mainLayout = QtWidgets.QGridLayout(self)

        self.toggleButton.setStyleSheet("QToolButton {border: none;}")
        self.toggleButton.setToolButtonStyle(QtCore.Qt.ToolButtonTextBesideIcon)
        self.toggleButton.setArrowType(QtCore.Qt.RightArrow)
        self.toggleButton.setText(title)
        self.toggleButton.setCheckable(True)
        self.toggleButton.setChecked(False)

        self.headerLine.setFrameShape(QtWidgets.QFrame.HLine)
        self.headerLine.setFrameShadow(QtWidgets.QFrame.Sunken)
        self.headerLine.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Maximum)

        # self.contentArea.setLayout(wd.QHBoxLayout())
        self.contentArea.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)

        # start out collapsed
        self.contentArea.setMaximumHeight(0)
        self.contentArea.setMinimumHeight(0)

        # let the entire widget grow and shrink with its content
        self.toggleAnimation.addAnimation(QtCore.QPropertyAnimation(self, b"minimumHeight"))
        self.toggleAnimation.addAnimation(QtCore.QPropertyAnimation(self, b"maximumHeight"))
        self.toggleAnimation.addAnimation(QtCore.QPropertyAnimation(self.contentArea, b"maximumHeight"))

        self.mainLayout.setVerticalSpacing(0)
        self.mainLayout.setContentsMargins(0, 0, 0, 0)

        row = 0
        self.mainLayout.addWidget(self.toggleButton, row, 0, 1, 1, QtCore.Qt.AlignLeft)
        self.mainLayout.addWidget(self.headerLine, row, 2, 1, 1)
        self.mainLayout.addWidget(self.contentArea, row + 1, 0, 1, 3)
        self.setLayout(self.mainLayout)

        self.toggleButton.toggled.connect(self.toggle)

    def set_layout(self, content_layout):
        layout = self.contentArea.layout()
        del layout
        self.contentArea.setLayout(content_layout)
        collapsed_height = self.sizeHint().height() - self.contentArea.maximumHeight()
        content_height = content_layout.sizeHint().height()
        for i in range(0, self.toggleAnimation.animationCount() - 1):
            section_animation = self.toggleAnimation.animationAt(i)
            section_animation.setDuration(self.animationDuration)
            section_animation.setStartValue(collapsed_height)
            section_animation.setEndValue(collapsed_height + content_height)
        content_animation = self.toggleAnimation.animationAt(self.toggleAnimation.animationCount() - 1)
        content_animation.setDuration(self.animationDuration)
        content_animation.setStartValue(0)
        content_animation.setEndValue(content_height)

    def toggle(self, collapsed):
        if collapsed:
            self.toggleButton.setArrowType(QtCore.Qt.DownArrow)
            self.toggleAnimation.setDirection(QtCore.QAbstractAnimation.Forward)
        else:
            self.toggleButton.setArrowType(QtCore.Qt.RightArrow)
            self.toggleAnimation.setDirection(QtCore.QAbstractAnimation.Backward)
        self.toggleAnimation.start()



"""Module for detecting Qt version."""
from __future__ import annotations

from ..qtpy.qt_compat import QT_API

QT_VERSION: str | None = None
if QT_API == "PySide6":
    from PySide6.QtCore import qVersion  # type: ignore

    QT_VERSION = qVersion()
elif QT_API == "PyQt6":
    from PyQt6.QtCore import qVersion  # type: ignore

    QT_VERSION = qVersion()
elif QT_API == "PyQt5":
    from PyQt5.QtCore import qVersion  # type: ignore

    QT_VERSION = qVersion()
elif QT_API == "PySide2":
    from PySide2.QtCore import qVersion  # type: ignore

    QT_VERSION = qVersion()



import pyqtgraph as pg
from qtpy import QtCore, QtWidgets

import argparse
import importlib
import logging
from platformdirs import user_config_dir, user_cache_dir
import sys

from .config.appearance import setup_appearance
from .config import Config, DataWidgets, SpectralLineData, Cache
from .io.viewer_data import add_unit_aliases
from .utils.params import LocalFile

from .widgets.MainWindow import MainWindow

logger = logging.getLogger(__name__)

ORGANIZATION_NAME = 'FRESCO'
APPLICATION_NAME = __package__.split('.')[0].capitalize()

CONFIG_DIR = user_config_dir(APPLICATION_NAME)
CACHE_DIR = user_cache_dir(APPLICATION_NAME)


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-v', '--verbosity', action='count', default=0)
    parser.add_argument('--purge', action='store_true')

    args = parser.parse_args()

    # configure logging parameters
    if args.verbosity == 0:
        level = logging.WARNING
    elif args.verbosity == 1:
        level = logging.INFO
    else:
        level = logging.DEBUG
    logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s', level=level)

    # configure pyqtgraph options
    pg.setConfigOption('imageAxisOrder', 'row-major')

    # read local config files
    local_files: dict[str, LocalFile] = {
        'config': LocalFile(CONFIG_DIR, filename='config.yml', full_name='General GUI settings'),
        'cache': LocalFile(CACHE_DIR, full_name='Cache', auto_backup=False),
        'widgets': LocalFile(CONFIG_DIR, filename='data_widgets.yml', full_name='Data viewer configuration'),
        'lines': LocalFile(CONFIG_DIR, filename='spectral_lines.yml', full_name='List of spectral lines'),
    }

    if args.purge:
        settings = QtCore.QSettings(ORGANIZATION_NAME, APPLICATION_NAME)
        settings.clear()
        settings.sync()
        for f in local_files.values():
            f.delete()  # safe delete

    config = Config.read_user_params(local_files['config'], default='config.yml')

    # register unit aliases
    add_unit_aliases(config.data.enabled_unit_aliases)

    # "discover" and "register" plugins
    plugins = []
    undiscovered_plugins = []
    for plugin_name in config.plugins:
        try:
            plugins.append(importlib.import_module("specvizitor.plugins." + plugin_name).Plugin())
        except ModuleNotFoundError:
            logger.warning(f'Plugin not found: {plugin_name}')
            undiscovered_plugins.append(plugin_name)

    config.plugins = [plugin_name for plugin_name in config.plugins if plugin_name not in undiscovered_plugins]
    config.save()

    exit_code = MainWindow.EXIT_CODE_REBOOT
    while exit_code == MainWindow.EXIT_CODE_REBOOT:
        # start the application
        app = QtWidgets.QApplication(sys.argv)
        app.setOrganizationName(ORGANIZATION_NAME)
        app.setApplicationName(APPLICATION_NAME)
        logger.info("Application started")

        # set up the GUI appearance
        setup_appearance(cfg=config.appearance)

        # create the main window
        window = MainWindow(config=config,
                            cache=Cache.read_user_params(local_files['cache']),
                            widget_cfg=DataWidgets.read_user_params(local_files['widgets'], default='data_widgets.yml'),
                            spectral_lines=SpectralLineData.read_user_params(local_files['lines'],
                                                                             default='spectral_lines.yml'),
                            plugins=plugins)
        window.show()

        exit_code = app.exec_()
        app = None
        logger.info("Application closed")

    sys.exit(exit_code)


if __name__ == '__main__':
    main()



import numpy as np
import pyqtgraph as pg
from pyqtgraph.dockarea.Container import StackedWidget
from pyqtgraph.dockarea.Dock import Dock
from qtpy import QtCore

import logging

from specvizitor.io.catalog import Catalog
from specvizitor.plugins.plugin_core import PluginCore

from specvizitor.widgets.ViewerElement import ViewerElement
from specvizitor.widgets.Image2D import Image2D

logger = logging.getLogger(__name__)


class Plugin(PluginCore):
    def override_widget_configs(self, widgets: dict[str, ViewerElement]):
        pass

    def update_docks(self, docks: dict[str, Dock], cat_entry: Catalog | None = None):
        rgb_docks = {title: d for title, d in docks.items() if "RGB" in title}

        rgb_stack = self.get_dock_stack(rgb_docks)

        if rgb_stack:
            self.fix_dock_stack_labels(rgb_stack)
            if cat_entry:
                self.raise_rgb_dock(rgb_stack, cat_entry)

    @staticmethod
    def raise_rgb_dock(rgb_stack: StackedWidget, cat_entry: Catalog):
        try:
            field = cat_entry.get_col('FIELD')
        except KeyError:
            return

        for i in range(rgb_stack.count()):
            d: Dock = rgb_stack.widget(i)
            if field in d.name():
                d.raiseDock()
                break

    def update_active_widgets(self, widgets: dict[str, ViewerElement], cat_entry: Catalog | None = None):

        spec_2d_stack: Image2D | None = widgets.get('Spectrum 2D [Stack]')
        spec_2d_beams: tuple[Image2D | None, ...] = tuple(widgets.get(f'Spectrum 2D [{label}]') for label in
                                                          ['Module A', 'Module B', 'R1, A', 'R1, B', 'R2, A', 'R2, B'])

        spec_2d_arr = tuple(spec_2d for spec_2d in (spec_2d_stack,) + spec_2d_beams if spec_2d is not None)

        # mark emission lines in the 2D spectra
        if cat_entry:
            self.mark_emline_from_cat(spec_2d_arr, cat_entry)

    def mark_emline_from_cat(self, spec_2d_arr: tuple[Image2D], cat_entry: Catalog):
        coords = self.get_emline_coords(cat_entry)
        for i, c in enumerate(coords):
            # add "-0.5" assuming that X_IMAGE and Y_IMAGE are calculated from Source-Extractor
            self.mark_emline_in_spec_2d(c[0]-0.5, c[1]-0.5, spec_2d_arr, zoom=True if i == 0 else False,
                                        color='r' if i == 0 else 'white')

    @staticmethod
    def get_emline_coords(cat_entry: Catalog) -> list[tuple[float, float]]:
        coords_array = []
        coord_colnames = [('X_IMAGE', 'Y_IMAGE')] + list((f'X_IMAGE_{i}', f'Y_IMAGE_{i}') for i in range(2, 11))

        for i, cname_pair in enumerate(coord_colnames):
            try:
                x, y = cat_entry.get_col(cname_pair[0]), cat_entry.get_col(cname_pair[1])
            except KeyError:
                break
            else:
                # skip masked elements
                if np.ma.is_masked(x) or np.ma.is_masked(y):
                    continue

                # convert line coordinates to float
                try:
                    line_coords = (float(x), float(y))
                except TypeError as msg:
                    logger.error(msg)
                else:
                    coords_array.append(line_coords)

        return coords_array

    @staticmethod
    def mark_emline_in_spec_2d(x0: float, y0: float, spec_2d_arr: tuple[Image2D], zoom=False, color='r'):
        zoom_level = 25
        dx, dy = 5, 5
        x1, x2 = x0 - zoom_level * dx, x0 + zoom_level * dx

        for spec_2d in spec_2d_arr:
            pen = pg.mkPen(color, width=2, style=QtCore.Qt.DashLine)

            spec_2d.register_item(pg.PlotCurveItem([x0 - dx, x0 + dx], [y0, y0], pen=pen))
            spec_2d.register_item(pg.PlotCurveItem([x0, x0], [y0 - dy, y0 + dy], pen=pen))

            if zoom:
                yrange = (0., float(spec_2d.data.shape[0]))
                spec_2d.set_default_range(xrange=(x1, x2), yrange=yrange, apply_qtransform=True)



from dataclasses import dataclass, field
from typing import Any

from ..utils.params import Params


@dataclass
class Catalogue:
    filename: str | None = None
    translate: dict[str, list[str]] = field(default_factory=dict)


@dataclass
class Image:
    filename: str
    wcs_source: str | None = None
    loader: str = 'auto'
    loader_params: dict[str, Any] = field(default_factory=dict)


@dataclass
class DataSources:
    dir: str = '.'
    images: dict[str, Image] = field(default_factory=dict)
    id_pattern: str = r'\d+'
    enabled_unit_aliases: dict[str, list[str]] = field(default_factory=dict)


@dataclass
class Appearance:
    theme: str = 'dark'
    antialiasing: bool = False
    viewer_spacing: int = 5
    viewer_margins: int = 5


@dataclass
class InspectionResults:
    default_flags: list[str] = field(default_factory=list)


@dataclass
class DataViewer:
    redshift_step: float = 0.01
    redshift_small_step: float = 0.0005


@dataclass
class Config(Params):
    appearance: Appearance = field(default_factory=Appearance)
    catalogue: Catalogue = field(default_factory=Catalogue)
    data: DataSources = field(default_factory=DataSources)
    data_viewer: DataViewer = field(default_factory=DataViewer)
    inspection_results: InspectionResults = field(default_factory=InspectionResults)
    plugins: list[str] = field(default_factory=list)



from dataclasses import dataclass, field

from ..utils.params import Params


@dataclass
class SpectralLineData(Params):
    wave_unit: str = "angstrom"
    repr: str = "vacuum"
    wavelengths: dict[str, float] = field(default_factory=dict)



from .AbstractWidget import AbstractWidget
from .ColorBar import ColorBar
from .FileBrowser import FileBrowser
from .ParamTable import ParamTable
from .Section import Section
from .MyViewBox import MyViewBox
from .MyQTextEdit import MyQTextEdit



from qtpy import QtWidgets

import abc

from ..qt_tools import get_widgets


class QtAbcMeta(type(QtWidgets.QWidget), type(abc.ABC)):
    pass


class AbstractWidget(QtWidgets.QWidget, abc.ABC, metaclass=QtAbcMeta):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.init_ui()
        self.set_layout()
        self.populate()

    def set_geometry(self, spacing: int, margins: int | tuple[int, int, int, int]):
        self.layout().setSpacing(spacing)

        if isinstance(margins, int):
            margins = tuple(margins for _ in range(4))
        self.layout().setContentsMargins(*margins)

    @abc.abstractmethod
    def init_ui(self):
        pass

    @abc.abstractmethod
    def set_layout(self):
        pass

    @abc.abstractmethod
    def populate(self):
        pass

    def repopulate(self):
        for widget in get_widgets(self.layout()):
            self.layout().removeWidget(widget)
        self.populate()



import pyqtgraph as pg


class MyViewBox(pg.ViewBox):
    def updateViewLists(self):
        try:
            self.window()
        except RuntimeError:  ## this view has already been deleted; it will probably be collected shortly.
            return

        nv = list(pg.ViewBox.NamedViews.values())

        if self in nv:
            nv.remove(self)

        if self.menu is not None:
            self.menu.setViewList(nv)

        for ax in [0, 1]:
            link = self.state['linkedViews'][ax]
            if isinstance(link, str):  ## axis has not been linked yet; see if it's possible now
                for v in nv:
                    if link == v.name:
                        self.linkView(ax, v)



import pyqtgraph as pg
from pgcolorbar.colorlegend import ColorLegendItem
import numpy as np

from .MyViewBox import MyViewBox


class ColorBar(ColorLegendItem):
    def __init__(self, *args, **kwargs):
        pg.ViewBox = MyViewBox
        super().__init__(*args, **kwargs)

    def _calcHistogramRange(self, imgArr, step='auto', targetImageSize=200):
        if imgArr is None or imgArr.size == 0 or np.all(~np.isfinite(imgArr)):
            return None, None

        mn, mx = self.getLevels()

        return mn, mx



import numpy as np
from platformdirs import user_config_dir
from qtpy import QtWidgets, QtCore

from abc import abstractmethod
import logging

from ..config import config
from ..config.spectral_lines import SpectralLineData
from ..io.catalog import Catalog, cat_browser
from ..io.viewer_data import data_browser
from ..utils.logs import qlog
from ..utils.widgets import AbstractWidget, FileBrowser, Section, ParamTable

logger = logging.getLogger(__name__)


class SettingsWidget(AbstractWidget):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.layout().setAlignment(QtCore.Qt.AlignTop)
        self.setFixedHeight(500)

    @abstractmethod
    def collect(self) -> bool:
        pass

    @abstractmethod
    def accept(self):
        pass


class AppearanceWidget(SettingsWidget):
    appearance_changed = QtCore.Signal(bool)

    THEMES = ('light', 'dark')

    def __init__(self, cfg: config.Appearance, parent=None):
        self.cfg = cfg

        self._theme_label: QtWidgets.QLabel | None = None
        self._theme_combobox: QtWidgets.QComboBox | None = None
        self._antialiasing_checkbox: QtWidgets.QCheckBox | None = None

        self._new_theme: str | None = None
        self._new_antialiasing: bool | None = None

        self._theme_changed: bool = False
        self._appearance_changed: bool = False

        super().__init__(parent=parent)

    def init_ui(self):
        self._theme_label = QtWidgets.QLabel("Theme:", self)
        self._theme_label.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)

        self._theme_combobox = QtWidgets.QComboBox(self)
        self._theme_combobox.addItems(self.THEMES)
        self._theme_combobox.setCurrentIndex(self.THEMES.index(self.cfg.theme))
        self._theme_combobox.setFixedWidth(200)

        self._antialiasing_checkbox = QtWidgets.QCheckBox("Antialiasing", self)
        self._antialiasing_checkbox.setChecked(self.cfg.antialiasing)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setSpacing(20)

    def populate(self):
        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.setAlignment(QtCore.Qt.AlignLeft)
        sub_layout.addWidget(self._theme_label)
        sub_layout.addWidget(self._theme_combobox)

        self.layout().addLayout(sub_layout)

        self.layout().addWidget(self._antialiasing_checkbox)

    def collect(self) -> bool:
        self._new_theme = self._theme_combobox.currentText()
        self._new_antialiasing = self._antialiasing_checkbox.isChecked()

        self._theme_changed = self._new_theme != self.cfg.theme
        if self._theme_changed or self._new_antialiasing != self.cfg.antialiasing:
            self._appearance_changed = True
        else:
            self._appearance_changed = False

        return True

    def accept(self):
        self.cfg.theme = self._new_theme
        self.cfg.antialiasing = self._new_antialiasing

        if self._appearance_changed:
            self.appearance_changed.emit(self._theme_changed)


def column_aliases_table_factory(translate: dict[str, list[str]], parent=None) -> ParamTable:
    header = ['Column', 'Aliases']
    data = [[cname, ','.join(cname_aliases)] for cname, cname_aliases in translate.items()]
    regex_pattern = [r'^\s*$', r'^\s*$']
    is_unique = [True, False]

    return ParamTable(header=header, data=data, name='Column Aliases', regex_pattern=regex_pattern,
                      is_unique=is_unique, remember_deleted=False, parent=parent)


class CatalogueWidget(SettingsWidget):
    data_requested = QtCore.Signal()
    catalog_changed = QtCore.Signal(object)

    def __init__(self, cat: Catalog, cfg: config.Catalogue, parent=None):
        self._old_cat = cat
        self.cfg = cfg

        self._new_cat: Catalog | None = None
        self._new_cat_filename: str | None = None
        self._new_translate = None

        self._catalog_changed = False
        self._aliases_changed = False

        self._browser: FileBrowser | None = None
        self._aliases_section: Section | None = None
        self._aliases_table: ParamTable | None = None

        super().__init__(parent=parent)

    def init_ui(self):
        self._browser = cat_browser(self.cfg.filename, title='Filename:', parent=self)

        self._aliases_section = Section("Column aliases", parent=self)
        self._aliases_table = column_aliases_table_factory(self.cfg.translate if self.cfg.translate else {},
                                                           parent=self)

        self.data_requested.connect(self._aliases_table.collect)
        self._aliases_table.data_collected.connect(self.save_aliases)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    def populate(self):
        self.layout().addWidget(self._browser)

        sub_layout = QtWidgets.QVBoxLayout()
        sub_layout.addWidget(self._aliases_table)
        self._aliases_section.set_layout(sub_layout)
        self.layout().addWidget(self._aliases_section)

    def collect(self) -> bool:
        self.data_requested.emit()

        if self._browser.is_filled():
            self._new_cat_filename = self._browser.path
            self._catalog_changed = self.cfg.filename is None or self._new_cat_filename != self.cfg.filename
        else:
            self._new_cat_filename = None
            self._catalog_changed = self.cfg.filename is not None

        if self._catalog_changed:
            if not self._browser.exists(verbose=True):
                return False

            if self._new_cat_filename is not None:
                self._new_cat = Catalog.read(self._new_cat_filename, translate=self._new_translate)
                if not self._new_cat:
                    return False
            else:
                self._new_cat = None

            return True

        if self._aliases_changed and self._old_cat is not None:
            self._new_cat = self._old_cat.update_translate(self._new_translate)
            if self._new_cat is None:
                return False

        return True

    @QtCore.Slot(list)
    def save_aliases(self, table_data: list[tuple[str, str]]):
        translate = {}
        for alias in table_data:
            translate[alias[0]] = alias[1].split(',')

        if self.cfg.translate.keys() == translate.keys() and list(self.cfg.translate.values()) == list(translate.values()):
            self._aliases_changed = False
        else:
            self._aliases_changed = True

        self._new_translate = translate

    def accept(self):
        self.cfg.translate = self._new_translate

        if self._catalog_changed:
            self.cfg.filename = self._new_cat_filename

        if self._catalog_changed or (self._aliases_changed and self._old_cat is not None):
            self.catalog_changed.emit(self._new_cat)


def image_table_factory(images: dict[str, config.Image], parent=None) -> ParamTable:
    header = ['Label', 'Path', 'WCS Source']
    data = [[label, img.filename, img.wcs_source] for label, img in images.items()]
    regex_pattern = [r'^\s*$', r'^\s*$', None]
    is_unique = [True, False, False]
    is_browser = [False, True, True]

    return ParamTable(header=header, data=data, name='Image', regex_pattern=regex_pattern,
                      is_unique=is_unique, remember_deleted=False, is_browser=is_browser, parent=parent)


class DataSourceWidget(SettingsWidget):
    data_requested = QtCore.Signal()
    images_changed = QtCore.Signal()

    def __init__(self, cfg: config.DataSources, parent=None):
        self.cfg = cfg

        self._browser: FileBrowser | None = None
        self._images_section: Section | None = None
        self._image_table: ParamTable | None = None

        self._new_dir: str | None = None
        self._new_images: dict[config.Image] | None = None

        self._images_changed: bool = False

        super().__init__(parent=parent)

    def init_ui(self):
        self._browser = data_browser(self.cfg.dir, title="Directory:", parent=self)

        self._images_section = Section("Images", parent=self)
        self._image_table = image_table_factory(self.cfg.images, self)

        self.data_requested.connect(self._image_table.collect)
        self._image_table.data_collected.connect(self.save_images)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    def populate(self):
        self.layout().addWidget(self._browser)

        sub_layout = QtWidgets.QVBoxLayout()
        sub_layout.addWidget(self._image_table)
        self._images_section.set_layout(sub_layout)
        self.layout().addWidget(self._images_section)

    def collect(self) -> bool:
        self.data_requested.emit()

        if not self._browser.is_filled(verbose=True) or not self._browser.exists(verbose=True):
            return False

        self._new_dir = self._browser.path
        return True

    @QtCore.Slot(list)
    def save_images(self, table_data: list[tuple[str, str, str]]):
        images = {}
        for img in table_data:
            images[img[0]] = config.Image(filename=img[1], wcs_source=img[2] if img[2] else None)

        old_data = [[label, img.filename, img.wcs_source] for label, img in self.cfg.images.items()] if self.cfg.images else []
        self._images_changed = True if old_data != table_data else False

        self._new_images = images

    def accept(self):
        self.cfg.dir = self._new_dir
        self.cfg.images = self._new_images

        if self._images_changed:
            self.images_changed.emit()


def spectral_lines_table_factory(wavelengths: dict[str, float], parent=None) -> ParamTable:
    header = ['Line', 'Wavelength']
    data = [[line, str(wave)] for line, wave in wavelengths.items()]
    regex_pattern = [r'^\s*$', r'(?!([0-9]+([.][0-9]*)?|[.][0-9]+)$)']
    is_unique = [True, False]

    return ParamTable(header=header, data=data, name='Spectral Line', regex_pattern=regex_pattern,
                      is_unique=is_unique, remember_deleted=False, parent=parent)


class DataViewerWidget(SettingsWidget):
    data_requested = QtCore.Signal()
    spectral_lines_changed = QtCore.Signal()

    def __init__(self, cfg: SpectralLineData, parent=None):
        self.cfg = cfg

        self._lines_section: Section | None = None
        self._lines_table: ParamTable | None = None

        self._new_wavelengths: dict[str, float] | None = None

        self._wavelengths_changed: bool = False

        super().__init__(parent=parent)

    def init_ui(self):
        self._lines_section = Section("Spectral lines", parent=self)
        self._lines_table = spectral_lines_table_factory(self.cfg.wavelengths, self)

        self.data_requested.connect(self._lines_table.collect)
        self._lines_table.data_collected.connect(self.save_lines)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    def populate(self):
        sub_layout = QtWidgets.QVBoxLayout()
        sub_layout.addWidget(self._lines_table)
        self._lines_section.set_layout(sub_layout)
        self.layout().addWidget(self._lines_section)

    def collect(self) -> bool:
        self.data_requested.emit()
        return True

    @QtCore.Slot(list)
    def save_lines(self, table_data: list[tuple[str, str]]):
        wavelengths = {}
        for line in table_data:
            wavelengths[line[0]] = float(line[1])

        self._wavelengths_changed = False
        if not self.cfg.wavelengths.keys() == wavelengths.keys():
            self._wavelengths_changed = True
        elif not np.isclose(np.fromiter(wavelengths.values(), dtype=float),
                            np.fromiter(self.cfg.wavelengths.values(), dtype=float)).all():
            self._wavelengths_changed = True

        self._new_wavelengths = wavelengths

    def accept(self):
        self.cfg.wavelengths = self._new_wavelengths

        if self._wavelengths_changed:
            self.spectral_lines_changed.emit()


class Settings(QtWidgets.QDialog):
    appearance_changed = QtCore.Signal()
    catalogue_changed = QtCore.Signal(object)
    data_source_changed = QtCore.Signal()
    spectral_lines_changed = QtCore.Signal()

    restart_requested = QtCore.Signal()

    def __init__(self, cat: Catalog, cfg: config.Config, spectral_lines: SpectralLineData, parent=None):
        self._old_cat = cat
        self._cfg = cfg
        self._spectral_lines = spectral_lines

        self._restart_required: bool = False

        self._tab_widget: QtWidgets.QTabWidget | None = None
        self._tabs: dict[str, SettingsWidget] | None = None

        self._info_label: QtWidgets.QLabel | None = None
        self._button_box: QtWidgets.QDialogButtonBox | None = None

        super().__init__(parent)
        self.setWindowTitle("Settings")

        self.init_ui()
        self.set_layout()
        self.populate()

    def create_tabs(self):
        self._tabs = {'Appearance': AppearanceWidget(self._cfg.appearance, self),
                      'Catalogue': CatalogueWidget(self._old_cat, self._cfg.catalogue, self),
                      'Data Source': DataSourceWidget(self._cfg.data, self),
                      'Data Viewer': DataViewerWidget(self._spectral_lines, self)}
        self._tabs['Appearance'].appearance_changed.connect(self._appearance_changed_action)
        self._tabs['Catalogue'].catalog_changed.connect(self.catalogue_changed.emit)
        self._tabs['Data Source'].images_changed.connect(self.data_source_changed.emit)
        self._tabs['Data Viewer'].spectral_lines_changed.connect(self.spectral_lines_changed.emit)

    def add_tabs(self):
        for name, t in self._tabs.items():
            self._tab_widget.addTab(t, name)

    def init_ui(self):
        self._tab_widget = QtWidgets.QTabWidget(self)

        self.create_tabs()
        self.add_tabs()

        self._info_label = QtWidgets.QLabel(f"Advanced settings: {user_config_dir('specvizitor')}", self)
        self._info_label.setTextInteractionFlags(QtCore.Qt.TextSelectableByMouse)

        # add OK/Cancel buttons
        self._button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel, self)

        self._button_box.accepted.connect(self.accept)
        self._button_box.rejected.connect(self.reject)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().setSpacing(20)

    def populate(self):
        self.layout().addWidget(self._tab_widget)
        self.layout().addWidget(self._info_label)
        self.layout().addWidget(self._button_box)

    @QtCore.Slot(bool)
    def _appearance_changed_action(self, theme_changed: bool):
        self._restart_required = self._restart_required or theme_changed
        self.appearance_changed.emit()

    @qlog
    def collect(self) -> bool:
        for t in self._tabs.values():
            if not t.collect():
                return False
        return True

    def accept(self):
        self._restart_required = False

        if not self.collect():
            return

        for t in self._tabs.values():
            t.accept()

        self._cfg.save()
        self._spectral_lines.save()

        restart_requested = False
        if self._restart_required:
            msg_box = QtWidgets.QMessageBox(self)
            ans = msg_box.question(self, '', f"Restart required to apply changes. Restart now?",
                                   msg_box.Yes | msg_box.No)
            if ans == msg_box.Yes:
                restart_requested = True

        super().accept()

        if restart_requested:
            self.restart_requested.emit()



from astropy.wcs import WCS
import numpy as np
from qtpy import QtCore, QtGui, QtWidgets


def get_widgets(layout: QtWidgets.QLayout) -> list[QtWidgets.QWidget]:
    widgets = []

    index = layout.count() - 1
    while index >= 0:
        widget, sub_layout = layout.itemAt(index).widget(), layout.itemAt(index).layout()
        if widget is not None:
            widgets.append(widget)
        elif layout is not None:
            widgets.extend(get_widgets(sub_layout))
        index -= 1

    return widgets


def get_qtransform_from_wcs(w: WCS) -> np.ndarray:
    transformation_matrix = np.zeros((3, 3))
    transformation_matrix[:2, :2] = w.pixel_scale_matrix
    transformation_matrix[2, :2] = w.wcs.crval
    transformation_matrix[2, 2] = 1.0

    # in pyqtgraph, the first pixel spans the [0, 1] range whereas in WCS we assume it's [0.5, 1.5]
    qtransform1 = QtGui.QTransform.fromTranslate(0.5, 0.5)

    # bring pixel values to the reference point
    qtransform2 = QtGui.QTransform.fromTranslate(-w.wcs.crpix[0], -w.wcs.crpix[1])

    qtransform3 = QtGui.QTransform()
    qtransform3.setMatrix(*transformation_matrix.flatten())

    return qtransform1 * qtransform2 * qtransform3


def safe_disconnect(signal: QtCore.Signal):
    try:
        signal.disconnect()
    except TypeError:
        pass



from astropy.table import Table
import astropy.units as u
import numpy as np
import pyqtgraph as pg
from pyqtgraph.dockarea.Container import StackedWidget
from pyqtgraph.dockarea.Dock import Dock
from qtpy import QtGui, QtWidgets

from functools import partial
import logging

from specvizitor.io.catalog import Catalog
from specvizitor.plugins.plugin_core import PluginCore

from specvizitor.widgets.ViewerElement import ViewerElement
from specvizitor.widgets.Image2D import Image2D
from specvizitor.widgets.Plot1D import Plot1D

logger = logging.getLogger(__name__)


class Plugin(PluginCore):
    LM_NAME = 'Line Map {}'

    def __init__(self):
        super().__init__()

        self._shortcuts_added = False

    def override_widget_configs(self, widgets: dict[str, ViewerElement]):
        """Override configurations of line maps.
        """
        lm: Image2D

        i = 1
        while widgets.get(self.LM_NAME.format(i)):
            lm = widgets.get(self.LM_NAME.format(i))
            if i > 1:
                lm.cfg.relative_to = self.LM_NAME.format(i - 1)
                lm.cfg.position = 'below'
            lm.cfg.data.loader_params['extver_index'] = i - 1  # EXTVER indexing starts at 0
            i += 1

    def update_docks(self, docks: dict[str, Dock], cat_entry: Catalog | None = None):
        """Add keyboard shortcuts to switch between line maps.
        """

        i = 1
        lm_docks = {}

        while docks.get(self.LM_NAME.format(i)):
            title = self.LM_NAME.format(i)
            lm_docks[title] = docks[title]
            i += 1

        lm_stack = self.get_dock_stack(lm_docks)

        if lm_stack:
            self.fix_dock_stack_labels(lm_stack)
            if cat_entry:
                self.raise_lm_dock(lm_stack, cat_entry)
            if not self._shortcuts_added:
                self._add_shortcuts(lm_docks, parent=lm_stack.widget(0).area.parent())

    @staticmethod
    def raise_lm_dock(lm_stack: StackedWidget, cat_entry: Catalog):
        lines = []
        for i in range(lm_stack.count()):
            d = lm_stack.widget(i)
            try:
                extver = d.widgets[0].meta["EXTVER"]
            except (TypeError, KeyError):
                lines.append(None)
            else:
                lines.append(extver)

        snr = []
        for i, line in enumerate(lines):
            try:
                snr_from_cat = cat_entry.get_col(f"sn_{line}")
            except KeyError:
                snr.append(0.)
            else:
                snr.append(snr_from_cat)

        i = int(np.argmax(snr))
        lm_stack.widget(i).raiseDock()


    def _add_shortcuts(self, lm_docks: dict[str, Dock], parent=None):
        QtWidgets.QShortcut('Up', parent, partial(self._switch_current_linemap, lm_docks, -1))
        QtWidgets.QShortcut('Down', parent, partial(self._switch_current_linemap, lm_docks, 1))
        self._shortcuts_added = True

    def _switch_current_linemap(self, lm_docks: dict[str, Dock], delta_index: int):
        lm_stack = self.get_dock_stack(lm_docks)
        if not lm_stack:
            return

        i = (lm_stack.currentIndex() + delta_index) % lm_stack.count()
        lm_stack.widget(i).raiseDock()

    def update_active_widgets(self, widgets: dict[str, ViewerElement], cat_entry: Catalog | None = None):
        spec_1d: Plot1D | None = widgets.get('Spectrum 1D')
        z_pdf: Plot1D | None = widgets.get('Redshift PDF')

        if spec_1d is not None:
            for w in widgets.values():
                if 'Spectrum 2D' in w.title:
                    self.transform_spec2d(w, spec_1d)

        if spec_1d is not None and z_pdf is not None:
            self.add_current_redshift_to_z_pdf(spec_1d, z_pdf)

        if spec_1d is not None:
            self.convert_spec1d_flux_unit_to_physical(spec_1d)
            spec_1d.setup_view(cat_entry)

    @staticmethod
    def transform_spec2d(spec_2d: Image2D, spec_1d: Plot1D):
        wave_unit = spec_1d._axes.x.unit
        if not wave_unit or not wave_unit.is_equivalent('AA'):
            wave_unit = 'AA'

        spec_2d._axes.x.unit = wave_unit
        spec_2d._axes.x.label = spec_1d._axes.x.label
        spec_2d.update_axis_labels()

        scale = (1 * u.Unit('micron')).to(wave_unit).value

        dlam = spec_2d.meta['CD1_1'] * scale
        crval = spec_2d.meta['CRVAL1'] * scale
        crpix = spec_2d.meta['CRPIX1']

        spec_2d._qtransform = QtGui.QTransform().translate(crval - dlam * crpix, 0).scale(dlam, 1)
        spec_2d.apply_qtransform(apply_to_default_range=True)

    @staticmethod
    def add_current_redshift_to_z_pdf(spec_1d: Plot1D, z_pdf: Plot1D):
        line = pg.InfiniteLine(spec_1d.redshift_slider.value, pen='m')
        spec_1d.redshift_slider.value_changed[float].connect(lambda z: line.setPos(z))
        z_pdf.register_item(line)

    @staticmethod
    def convert_spec1d_flux_unit_to_physical(spec_1d: Plot1D):
        t = spec_1d.data
        if not isinstance(t, Table):
            logger.error('Spectrum 1D data must be of the `astropy.table.Table` type')
            return

        # check that the `flux` and `flat` columns are in the table
        if 'flux' not in t.colnames:
            logger.error('Column not found: flux')
        if 'flat' not in t.colnames:
            logger.error('Column not found: flat')

        # check that the flux unit is correct
        if not t['flux'].unit or not t['flux'].unit.is_equivalent(u.Unit('ct / s')):
            logger.error('The input flux unit must be `ct/s`')
            return

        # convert fluxes to physical units
        flat = t['flat'].to('1e19 AA cm2 ct / erg')

        with np.errstate(divide='ignore'):
            scale = 1 / flat
        scale[scale == np.inf] = np.nan

        y_unit = (t['flux'] * scale).unit
        spec_1d._axes.y.unit = y_unit

        for label in ('flux', 'err', 'model'):
            plot_data_item = spec_1d.plot_data_items.get(label)
            if plot_data_item is None:
                continue

            x_data, _ = plot_data_item.getData()

            y_data = spec_1d.get_plot_data(spec_1d.cfg.plots[label].y)
            y_data *= scale
            y_data = spec_1d.apply_ydata_transform(y_data)

            t[label] = y_data * y_unit  # replace the original data
            plot_data_item.setData(x=x_data, y=y_data)



import qtpy.compat
from qtpy import QtCore, QtWidgets

import logging
import pathlib

from .AbstractWidget import AbstractWidget

logger = logging.getLogger(__name__)


class FileBrowser(AbstractWidget):
    textChanged = QtCore.Signal(str)

    OpenFile = 0
    # OpenFiles = 1
    OpenDirectory = 2
    SaveFile = 3

    def __init__(self, mode=OpenFile, filename_extensions='All files (*.*)', default_path=None,
                 title: str = '', title_width: int | None = None, button_text='Browse...',
                 line_edit_width=700, parent=None):

        self._browser_mode = mode
        self._filter = filename_extensions  # example: 'Images (*.png *.xpm *.jpg);;Text files (*.txt)'
        self._default_path = None if default_path is None else str(pathlib.Path(default_path).resolve())

        self._title = title
        self._title_width = title_width
        self._button_text = button_text
        self._line_edit_width = line_edit_width

        self._label: QtWidgets.QLabel | None = None
        self._line_edit: QtWidgets.QLineEdit | None = None
        self._button: QtWidgets.QPushButton | None = None

        super().__init__(parent=parent)

    def init_ui(self):
        self._label = QtWidgets.QLabel(self._title, self)
        if self._title_width is None:
            self._label.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        else:
            self._label.setFixedWidth(self._title_width)
        if not self._title:
            self._label.setVisible(False)

        self._line_edit = QtWidgets.QLineEdit(self)
        self._line_edit.setMinimumWidth(self._line_edit_width)
        self._line_edit.setText(self._default_path)

        self._button = QtWidgets.QPushButton(self._button_text, self)
        self._button.setFixedWidth(120)

        self._line_edit.textChanged.connect(self.textChanged.emit)
        self._button.clicked.connect(self._browse)

    def set_layout(self):
        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)

    def populate(self):
        self.layout().addWidget(self._label)
        self.layout().addWidget(self._line_edit)
        self.layout().addWidget(self._button)

    @property
    def path(self) -> str:
        return self._line_edit.text()

    def is_filled(self, verbose=False) -> bool:
        if not self.path:
            if verbose:
                logger.error('The field `{}` is empty'.format(self._title.rstrip(':')))
            return False
        return True

    def exists(self, verbose=False):
        if self._browser_mode == FileBrowser.SaveFile:
            path_to_check = pathlib.Path(self.path).parent
        else:
            path_to_check = pathlib.Path(self.path)

        if not path_to_check.exists():
            if verbose:
                if self._browser_mode == FileBrowser.OpenFile:
                    msg = 'The file `{}` does not exist'.format(path_to_check)
                elif self._browser_mode in (FileBrowser.OpenDirectory, FileBrowser.SaveFile):
                    msg = 'The directory `{}` does not exist'.format(path_to_check)
                else:
                    msg = 'The path `{}` does not exist'.format(path_to_check)
                logger.error(msg)
            return False

        return True

    def _browse(self):
        path = None

        if self._browser_mode == FileBrowser.OpenFile:
            path = qtpy.compat.getopenfilename(self, caption='Choose File', basedir=self._default_path,
                                               filters=self._filter)[0]

        # elif self._browser_mode == FileBrowser.OpenFiles:
        #     self._filepaths.extend(qtpy.compat.getopenfilenames(self, caption='Choose Files',
        #                                                                   basedir=self._default_path,
        #                                                                   filters=self._filter)[0])

        elif self._browser_mode == FileBrowser.OpenDirectory:
            path = qtpy.compat.getexistingdirectory(self, caption='Choose Directory', basedir=self._default_path)

        elif self._browser_mode == FileBrowser.SaveFile:
            path, extension = qtpy.compat.getsavefilename(self, caption='Save/Save As', basedir=self._default_path,
                                                          filters=self._filter)

        if path:
            self._line_edit.setText(path)



from qtpy import QtWidgets, QtCore

import logging

from ..io.catalog import Catalog
from ..io.inspection_data import InspectionData
from ..utils.widgets import AbstractWidget

from .TableColumns import TableColumns

logger = logging.getLogger(__name__)


class ObjectInfo(AbstractWidget):
    data_collected = QtCore.Signal(list)

    def __init__(self, parent=None):
        self.all_columns: dict[str, str] | None = None
        self.visible_columns: list[str] | None = None

        self._table: QtWidgets.QTableWidget | None = None
        self._table_items: dict[str, list[tuple[QtWidgets.QTableWidgetItem, QtWidgets.QTableWidgetItem]]] | None = None

        self._search_label: QtWidgets.QLabel | None = None
        self._search_lineedit: QtWidgets.QLineEdit | None = None
        self._display_options: QtWidgets.QPushButton | None = None

        super().__init__(parent=parent)
        self.setEnabled(False)
        self.setSizePolicy(QtWidgets.QSizePolicy.MinimumExpanding, QtWidgets.QSizePolicy.Preferred)

    def _create_table_items(self):
        if self.all_columns is None:
            self._table_items = {}
            return

        table_items = {}
        for cname, cname_annotated in self.all_columns.items():
            cname_item = QtWidgets.QTableWidgetItem(cname_annotated)
            value_item = QtWidgets.QTableWidgetItem('')

            table_items[cname] = cname_item, value_item

        self._table_items = table_items

    def _set_table_items(self):
        self._table.setRowCount(len(self._table_items))
        for i, row in enumerate(self._table_items.values()):
            self._table.setItem(i, 0, row[0])
            self._table.setItem(i, 1, row[1])

    @QtCore.Slot(object)
    def update_table_items(self, cat: Catalog | None):
        self.all_columns = cat.annotated_colnames if cat is not None else None

        self._create_table_items()
        self._set_table_items()
        if self.all_columns:
            self.update_visible_columns(list(self.all_columns.keys()))

    @QtCore.Slot()
    def update_view(self):
        for i, cname in enumerate(self._table_items.keys()):
            if cname in self.visible_columns and self._search_lineedit.text() in cname:
                self._table.showRow(i)
            else:
                self._table.hideRow(i)

    @QtCore.Slot(list)
    def update_visible_columns(self, visible_columns: list[str] | None):
        self.visible_columns = visible_columns
        self.update_view()

    def init_ui(self):
        self._table = QtWidgets.QTableWidget(self)
        self._table.setColumnCount(2)
        self._table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        # self._table.horizontalHeader().hide()
        self._table.horizontalHeader().setStretchLastSection(True)
        # self._table.horizontalHeader().setSectionResizeMode(0, QtWidgets.QHeaderView.ResizeToContents)
        self._table.setHorizontalHeaderLabels(('Key', 'Value'))
        self._table.verticalHeader().hide()
        self._table.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Fixed)

        self._search_label = QtWidgets.QLabel("Search:", self)
        self._search_label.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)
        self._search_lineedit = QtWidgets.QLineEdit(self)
        self._search_lineedit.setSizePolicy(QtWidgets.QSizePolicy.Ignored, QtWidgets.QSizePolicy.Maximum)
        self._search_lineedit.setMinimumWidth(50)
        self._display_options = QtWidgets.QPushButton("Columns...", self)
        self._display_options.setSizePolicy(QtWidgets.QSizePolicy.Maximum, QtWidgets.QSizePolicy.Maximum)

        self._search_lineedit.textChanged[str].connect(self.update_view)
        self._display_options.clicked.connect(self._display_options_action)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    def populate(self):
        self.layout().addWidget(self._table)

        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.addWidget(self._search_label)
        sub_layout.addWidget(self._search_lineedit)
        sub_layout.addWidget(self._display_options)

        self.layout().addLayout(sub_layout)

    @QtCore.Slot()
    def load_project(self):
        self.setEnabled(True)

    @QtCore.Slot(int, InspectionData, object)
    def load_object(self, _, __, cat_entry: Catalog | None):

        if cat_entry is None:
            for row in self._table_items.values():
                row[1].setText('')
            return

        for cname, row in self._table_items.items():
            try:
                value = cat_entry.get_col(cname)
                value = f'{value:.8f}' if isinstance(value, float) else str(value)
                row[1].setText(value)
            except KeyError as e:
                logger.warning(e)
                row[1].setText('')

    @QtCore.Slot()
    def collect(self):
        self.data_collected.emit(self.visible_columns)

    def _display_options_action(self):
        dialog = TableColumns(list(self.all_columns.keys()), self.visible_columns, parent=self)
        dialog.visible_columns_updated.connect(self.update_visible_columns)
        if dialog.exec():
            pass



from qtpy import QtCore, QtWidgets

from itertools import repeat
import logging

from ..io.inspection_data import InspectionData
from ..utils.widgets import ParamTable

logger = logging.getLogger(__name__)


def inspection_field_table_factory(review: InspectionData, parent=None) -> ParamTable:
    header = ['Field', 'Type']
    data = list(zip(review.flag_columns, repeat('boolean')))
    item_choices = [None, ('boolean',)]

    reserved_colnames = list(review.indices) + review.default_columns
    regex_pattern = [fr"^({'|'.join(reserved_colnames)}|\s*)$", None]

    is_unique = [True, False]

    return ParamTable(header=header, data=data, name='Inspection Field', item_choices=item_choices,
                      regex_pattern=regex_pattern, is_unique=is_unique, parent=parent)


class InspectionFieldEditor(QtWidgets.QDialog):
    data_requested = QtCore.Signal()
    inspection_fields_updated = QtCore.Signal(list, list, bool)

    def __init__(self, review: InspectionData, parent=None):
        self._review = review
        self._field_table: ParamTable | None = None
        self._set_as_default_checkbox: QtWidgets.QCheckBox | None = None
        self._button_box: QtWidgets.QDialogButtonBox | None = None

        self._new_fields: list[tuple[str, str]] | None = None
        self._is_deleted: list[bool] | None = None

        super().__init__(parent=parent)
        self.setWindowTitle("Edit Inspection Fields")

        self.init_ui()
        self.set_layout()
        self.populate()

    def init_ui(self):
        self._field_table = inspection_field_table_factory(self._review, self)
        self._set_as_default_checkbox = QtWidgets.QCheckBox('Set as default for new inspection files', self)

        self._button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel, self)

        self._button_box.accepted.connect(self.accept)
        self._button_box.rejected.connect(self.reject)

        self.data_requested.connect(self._field_table.collect)
        self._field_table.data_collected.connect(self.save_inspection_fields)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    def populate(self):
        self.layout().addWidget(self._field_table)
        self.layout().addWidget(self._set_as_default_checkbox)
        self.layout().addWidget(self._button_box)

    @QtCore.Slot(list, list)
    def save_inspection_fields(self, fields: list[tuple[str, str]], is_deleted: list[bool]):
        self._new_fields = fields
        self._is_deleted = is_deleted

    def accept(self):
        self._new_fields = None
        self._is_deleted = None
        self.data_requested.emit()

        warn_columns = []
        for i, old_name in enumerate(self._review.user_defined_columns):
            if self._is_deleted[i] and self._review.has_data(old_name):
                warn_columns.append(old_name)

        accept = True
        if warn_columns:
            msg_box = QtWidgets.QMessageBox(self)
            ans = msg_box.question(self, '', f"Data in the following column(s) will be deleted: {', '.join(warn_columns)}. "
                                             "Are you sure you want to proceed?", msg_box.Yes | msg_box.No)
            if ans == msg_box.No:
                accept = False

        if accept:
            self.inspection_fields_updated.emit(self._new_fields, self._is_deleted, self._set_as_default_checkbox.isChecked())
            super().accept()



from dataclasses import dataclass
import logging
import pathlib

from .io.catalog import Catalog
from .io.inspection_data import InspectionData

logger = logging.getLogger(__name__)


@dataclass
class AppData:

    output_path: pathlib.Path | None = None  # the path to the output (a.k.a. inspection) file
    cat: Catalog | None = None               # the catalogue
    review: InspectionData | None = None     # inspection results

    j: int = None  # the index of the current object

    def create(self, **kwargs):
        """ Initialize the inspection data object.
        """
        if self.cat is None:
            logger.error("Failed to initialize inspection data: the catalogue not loaded to the memory")
            return

        self.review = InspectionData.create(*[list(self.cat.get_col(ind)) for ind in self.cat.indices], **kwargs)
        logger.info(f"Project created (path: {self.output_path})")

    def read(self):
        """ Read the inspection file.
        """
        if self.output_path is None:
            logger.error("Failed to read the inspection file: the file path not specified")
            return

        self.review = InspectionData.read(self.output_path)
        logger.info(f"Project loaded (path: {self.output_path})")

    def save(self):
        """ Save inspection data to the output file.
        """
        if self.output_path is None:
            logger.error("Failed to save the inspection data: the output path not specified")
            return

        self.review.write(self.output_path)
        logger.info(f"Project saved (path: {self.output_path})")



from pyqtgraph.dockarea.Dock import Dock
from pyqtgraph.dockarea.DockArea import DockArea
from qtpy import QtWidgets, QtCore

from functools import partial
import logging
import time

from ..config import config, data_widgets
from ..config.data_widgets import DataWidgets
from ..config.spectral_lines import SpectralLineData
from ..io.catalog import Catalog
from ..io.inspection_data import InspectionData
from ..io.viewer_data import ViewerData
from ..plugins.plugin_core import PluginCore
from ..utils.widgets import AbstractWidget

from .ViewerElement import ViewerElement, LinkableItem, SliderItem
from .Image2D import Image2D
from .Plot1D import Plot1D
from .SmartSlider import SmartSlider
from .ViewerDataLoader import ViewerDataLoader
from .ItemLinker import ItemLinker, XAxisLinker, YAxisLinker, SliderLinker, ColorBarLinker

logger = logging.getLogger(__name__)


class DataViewer(AbstractWidget):
    project_loaded = QtCore.Signal()
    data_loaded = QtCore.Signal(int, InspectionData, object)
    object_loaded = QtCore.Signal()
    loading_aborted = QtCore.Signal()
    data_collected = QtCore.Signal(dict)

    redshift_requested = QtCore.Signal()
    redshift_changed = QtCore.Signal(float)
    redshift_obtained = QtCore.Signal(float)

    view_reset = QtCore.Signal(object)
    visibility_changed = QtCore.Signal(bool)
    spectral_lines_changed = QtCore.Signal()

    def __init__(self,
                 global_cfg: config.DataViewer,
                 data_cfg: config.DataSources,
                 widget_cfg: DataWidgets,
                 appearance: config.Appearance,
                 spectral_lines: SpectralLineData,
                 plugins: list[PluginCore],
                 parent=None):

        self._global_cfg = global_cfg
        self._data_cfg = data_cfg
        self._widget_cfg = widget_cfg
        self._appearance = appearance
        self._spectral_lines = spectral_lines
        self._plugins = plugins

        self._zen_mode_activated: bool = False

        self._data = ViewerData()
        self.open_images()

        self._widget_links: dict[LinkableItem, dict] = {item: dict() for item in LinkableItem}
        self._widget_linkers: dict[LinkableItem, ItemLinker] = dict()
        self._create_widget_linkers()

        self._worker: ViewerDataLoader | None = None
        self._lock: bool = False
        self._t_worker_start = None
        self._t_old_worker_start = None

        self.dock_area: DockArea | None = None
        self._added_docks: list[str] = []

        self.widgets: dict[str, ViewerElement] = {}
        self.docks: dict[str, Dock] = {}

        super().__init__(parent=parent)
        self.setEnabled(False)
        self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

        QtWidgets.QShortcut('Q', self, partial(self._change_redshift, -1))
        QtWidgets.QShortcut('W', self, partial(self._change_redshift, 1))

        QtWidgets.QShortcut('Shift+Q', self, partial(self._change_redshift, -1, True))
        QtWidgets.QShortcut('Shift+W', self, partial(self._change_redshift, 1, True))

    @property
    def active_widgets(self) -> dict[str, ViewerElement]:
        return {wt: w for wt, w in self.widgets.items() if w.data is not None}

    def _create_widget(self, wt: str, cfg: data_widgets.ViewerElement):
        constructor = None
        if isinstance(cfg, data_widgets.Image):
            constructor = Image2D
        elif isinstance(cfg, data_widgets.Plot1D):
            constructor = Plot1D
        if constructor is None:
            logger.error(f"Unknown widget configuration type: {type(cfg)}")

        kwargs = dict(
            appearance=self._appearance,
            spectral_lines=self._spectral_lines,
            parent=self
        )

        self.widgets[wt] = constructor(cfg=cfg, title=wt, **kwargs)
        self._connect_widget(wt)

    def _create_widgets(self):
        for wt in list(self.widgets):
            self._delete_widget(wt)

        for name, image_cfg in self._widget_cfg.images.items():
            self._create_widget(name, image_cfg)

        for name, plot_cfg in self._widget_cfg.plots.items():
            self._create_widget(name, plot_cfg)

        for plugin in self._plugins:
            plugin.override_widget_configs(self.widgets)

    def _delete_widget(self, wt: str):
        self._unlink_widget(wt)
        self._disconnect_widget(wt)

        w0 = self.widgets.pop(wt)

        w0.graphics_layout.clear()
        w0.deleteLater()

    def _connect_widget(self, wt: str):
        w0 = self.widgets[wt]

        self.project_loaded.connect(w0.load_project)
        self.data_loaded.connect(w0.load_object)
        self.view_reset.connect(w0.reset_view)
        self.spectral_lines_changed.connect(w0.update_spectral_lines)

        w0.object_loaded.connect(self._attach_widget)
        w0.object_destroyed.connect(self._detach_widget)
        w0.redshift_slider.save_button_clicked.connect(self._save_redshift)

    def _disconnect_widget(self, wt: str):
        w0 = self.widgets[wt]

        self.project_loaded.disconnect(w0.load_project)
        self.data_loaded.disconnect(w0.load_object)
        self.view_reset.disconnect(w0.reset_view)
        self.spectral_lines_changed.disconnect(w0.update_spectral_lines)

        w0.object_loaded.disconnect()
        w0.object_destroyed.disconnect()
        w0.redshift_slider.save_button_clicked.disconnect()

    def _create_widget_linkers(self):
        self._widget_linkers = {
            LinkableItem.XAXIS: XAxisLinker(),
            LinkableItem.YAXIS: YAxisLinker(),
            LinkableItem.COLORBAR: ColorBarLinker(),
            LinkableItem.S_SLIDER: SliderLinker(SliderItem.SMOOTHING),
            LinkableItem.S_REDSHIFT: SliderLinker(SliderItem.REDSHIFT)
        }

    def _link_widget(self, wt: str):
        w0 = self.widgets[wt]

        for w in self.active_widgets.values():
            for linked_item, linker in self._widget_linkers.items():
                linker.link(w0, w, self._widget_links[linked_item])
                linker.link(w, w0, self._widget_links[linked_item])

    def _unlink_widget(self, wt: str):
        w0 = self.widgets[wt]

        for w in self.widgets.values():
            for linked_item, linker in self._widget_linkers.items():
                linker.unlink(w0, w, self._widget_links[linked_item])
                linker.unlink(w, w0, self._widget_links[linked_item])

    def _close_dock(self, dt: str, dock: Dock | None = None):
        if dock is None:
            dock = self.docks[dt]

        container = dock.container()
        dock.close()

        if container and not isinstance(container, DockArea):
            self._close_dock(dt, container)
        else:
            self._added_docks.remove(dt)
            self.docks.pop(dt)
            return

    def _create_dock(self, dt: str):
        w0 = self.widgets[dt]
        self.docks[dt] = Dock(dt, widget=w0)
        self._add_dock(dt)

    def _add_dock(self, dt: str):
        dock, w0 = self.docks[dt], self.widgets[dt]
        if not w0.cfg.visible:
            return

        position = w0.cfg.position if w0.cfg.position is not None else 'bottom'
        relative_to = w0.cfg.relative_to if w0.cfg.relative_to in self._added_docks else None

        self.dock_area.addDock(dock=dock, position=position, relativeTo=relative_to)
        self._update_visibility(dt)
        self._added_docks.append(dt)

    def _create_docks(self):
        for dt in list(self.docks):
            self._close_dock(dt)

        for wt in self.widgets:
            self._create_dock(wt)

        for plugin in self._plugins:
            plugin.update_docks(self.docks)

    @QtCore.Slot()
    def reset_dock_layout(self):
        self._create_docks()

    @staticmethod
    def _clean_dock_layout(layout: dict):
        float_layout_cleaned = []
        for l in layout['float']:
            if l[0]['main'] is not None:
                float_layout_cleaned.append(l)

        layout['float'] = float_layout_cleaned

    @QtCore.Slot(dict)
    def restore_dock_layout(self, layout: dict):
        # fix for a pyqtgraph bug where empty layouts throw an error
        self._clean_dock_layout(layout)

        try:
            # set `extra` to None to catch an exception (KeyError) when adding extra docks not mentioned in `layout`
            self.dock_area.restoreState(layout, missing='ignore', extra=None)
            logger.info("Dock layout restored")

            for plugin in self._plugins:
                plugin.update_docks(self.docks)
        except (KeyError, ValueError, TypeError) as e:
            self.reset_dock_layout()
            logger.error(f"Failed to restore the dock layout: {e}")

    def init_ui(self):
        if self.dock_area is None:
            self.dock_area = DockArea(parent=self)

        self._create_widgets()
        self._create_docks()

    def set_layout(self):
        self.setLayout(QtWidgets.QGridLayout())
        self.set_geometry(spacing=0, margins=0)

    def populate(self):
        self.layout().addWidget(self.dock_area, 1, 1, 1, 1)

    @QtCore.Slot(DataWidgets)
    def update_viewer_configuration(self, viewer_cfg: DataWidgets):
        self._widget_cfg = viewer_cfg
        self.init_ui()

    @QtCore.Slot()
    def open_images(self):
        if self._data_cfg.images is None:
            return

        for img_label, img_cfg in self._data_cfg.images.items():
            self._data.open_image(filename=img_cfg.filename, loader=img_cfg.loader, wcs_source=img_cfg.wcs_source,
                                  **img_cfg.loader_params)

    @QtCore.Slot()
    def load_project(self):
        self._lock = True
        self.setEnabled(True)
        self.project_loaded.emit()

    @QtCore.Slot(int, InspectionData, object)
    def load_object(self, j: int, review: InspectionData, cat_entry: Catalog | None):
        if self._worker and self._worker.isRunning():
            self._worker.wait()

        self._t_worker_start = time.perf_counter()
        t_grace = self._get_t_grace()
        self._t_old_worker_start = self._t_worker_start

        self._worker = ViewerDataLoader(self.widgets, j, review, self._data, self._data_cfg, cat_entry, t_grace=t_grace)
        self.loading_aborted.connect(self._worker.abort)
        self._worker.finished.connect(self.finalize_loading)
        self._worker.start()

        if self._lock:
            self._worker.wait()

    @QtCore.Slot()
    def abort_loading(self):
        if self._worker and self._worker.isRunning():
            self._worker.finished.disconnect(self.finalize_loading)

        self.loading_aborted.emit()

    def _get_t_grace(self):
        dt = 1000 if self._t_old_worker_start is None else self._t_worker_start - self._t_old_worker_start
        t_grace = 0.35 if 0.10 < dt < 0.35 else 0.10
        return t_grace

    @QtCore.Slot()
    def finalize_loading(self):
        j, review, cat_entry = self._worker.j, self._worker.review, self._worker.cat_entry

        self.data_loaded.emit(j, review, cat_entry)

        for plugin in self._plugins:
            plugin.update_active_widgets(self.active_widgets, cat_entry=cat_entry)
            plugin.update_docks(self.docks, cat_entry=cat_entry)

        self.reset_view()

        self._lock = False
        self.object_loaded.emit()

    @QtCore.Slot(str)
    def _attach_widget(self, wt: str):
        w0 = self.widgets[wt]
        self._link_widget(wt)
        self.docks[wt].setTitle(w0.get_dock_title())
        logger.debug(f"`{wt}` attached")

    @QtCore.Slot(str)
    def _detach_widget(self, wt: str):
        w0 = self.widgets[wt]
        self._unlink_widget(wt)
        self.docks[wt].setTitle(w0.title)
        logger.debug(f"`{wt}` detached")

    def _get_active_redshift_slider(self) -> SmartSlider | None:
        for w in self.active_widgets.values():
            if w.redshift_slider.isVisible():
                return w.redshift_slider
        return None

    @QtCore.Slot()
    def reset_view(self):
        self.view_reset.emit(self._widget_links)

    @QtCore.Slot()
    def request_redshift(self):
        slider = self._get_active_redshift_slider()
        if slider is None:
            return

        self.redshift_requested.connect(slider.save_value)
        self.redshift_requested.emit()
        self.redshift_requested.disconnect()

    @QtCore.Slot(float)
    def _save_redshift(self, redshift: float):
        self.redshift_obtained.emit(redshift)

    def _change_redshift(self, n_steps: int, small_step: bool = False):
        slider = self._get_active_redshift_slider()
        if slider is None:
            return

        self.redshift_changed.connect(slider.set_value)

        z = slider.value
        step = self._global_cfg.redshift_small_step if small_step else self._global_cfg.redshift_step
        dz = n_steps * step * (1 + z)

        self.redshift_changed.emit(z + dz)
        self.redshift_changed.disconnect()

    @QtCore.Slot()
    def collect(self):
        self.data_collected.emit(self.dock_area.saveState())

    @QtCore.Slot(str)
    def take_screenshot(self, filename: str):
        self.grab().save(filename)

    @QtCore.Slot(bool)
    def enter_zen_mode(self, is_zen: bool):
        self._zen_mode_activated = is_zen
        for wt in self.widgets:
            self._update_visibility(wt)

    def _update_visibility(self, wt: str):
        w0 = self.widgets[wt]
        dock = self.docks[wt]

        if self._zen_mode_activated:
            dock.hideTitleBar()
        else:
            dock.showTitleBar()

        self.visibility_changed.connect(w0.update_visibility)
        self.visibility_changed.emit(self._zen_mode_activated)
        self.visibility_changed.disconnect(w0.update_visibility)

    @QtCore.Slot()
    def free_resources(self):
        self._data.close_all()



"""PyQtDarkTheme - A flat dark theme for PySide and PyQt.

- Repository: https://github.com/5yutan5/PyQtDarkTheme
- Documentation: https://pyqtdarktheme.readthedocs.io


License Information
===================

Material design icons
---------------------

All svg files in PyQtDarkTheme is from Material design icons(which uses an Apache 2.0 license).

- Author: Google
- Site: https://fonts.google.com/icons
- Source: https://github.com/google/material-design-icons
- License: Apache License Version 2.0 | https://www.apache.org/licenses/LICENSE-2.0.txt

Modifications made to each files to change the icon color and angle and remove svg namespace.

The current Material design icons license summary can be viewed at:
https://github.com/google/material-design-icons/blob/master/LICENSE


QDarkStyleSheet(Source code)
----------------------------

Qt stylesheets are originally fork of QDarkStyleSheet(MIT License).

- Author: Colin Duquesnoy
- Site: https://github.com/ColinDuquesnoy/QDarkStyleSheet
- Source: https://github.com/ColinDuquesnoy/QDarkStyleSheet
- License: MIT License | https://opensource.org/licenses/MIT

Modifications made to a file to change the style.

The current QDarkStyleSheet license summary can be viewed at:
https://github.com/ColinDuquesnoy/QDarkStyleSheet/blob/master/LICENSE.rst

"""
# Version of PyQtDarkTheme
QDARKTHEME_VERSION = "2.1.0"

from ._main import enable_hi_dpi, setup_theme, stop_sync
from ._style_loader import clear_cache, get_themes, load_palette, load_stylesheet



"""Package including resources.

**Warning**

This package created programmatically. All changes made in this file will be lost!
Created by the `PyQtDarkTheme/tools/build_styles`.

"""

from . import colors, palette, stylesheets, svg

THEMES = ("dark", "light", "auto")



from dataclasses import dataclass, field
from typing import Any

from ..utils.params import Params


@dataclass
class Limits:
    min: float | None = None
    max: float | None = None


@dataclass
class ColorBarLimits(Limits):
    type: str = 'minmax'


@dataclass
class Slider:
    visible: bool = True
    link_to: str | None = None

    min_value: float = 0
    max_value: float = 100
    step: float = 1
    default_value: float = 0

    catalog_name: str | None = None
    show_text_editor: bool = False
    n_decimal_places: int = 6

    show_save_button: bool = False


@dataclass
class ColorBar:
    visible: bool = True
    link_to: str | None = None

    limits: ColorBarLimits = field(default_factory=ColorBarLimits)


@dataclass
class SpectralLines:
    visible: bool = False
    color: str | None = None


@dataclass
class LinePlot:
    x: str
    y: str
    color: str | None = None
    hide_label: bool = False


@dataclass
class Axis:
    visible: bool = True
    link_to: str | None = None

    unit: str | None = None
    scale: str = 'linear'
    label: str | None = None
    limits: Limits = field(default_factory=Limits)


@dataclass
class DataElement:
    source: str | None = None
    filename: str | None = None
    loader: str = 'auto'
    loader_params: dict[str, Any] = field(default_factory=dict)


@dataclass
class ViewerElement:
    visible: bool = True
    position: str | None = None
    relative_to: str | None = None
    dock_title_fmt: str = 'short'

    data: DataElement = field(default_factory=DataElement)

    x_axis: Axis = field(default_factory=Axis)
    y_axis: Axis = field(default_factory=Axis)

    spectral_lines: SpectralLines = field(default_factory=SpectralLines)
    smoothing_slider: Slider = field(default_factory=lambda: Slider(max_value=3, step=0.05))
    redshift_slider: Slider = field(default_factory=lambda: Slider(visible=False, max_value=10, step=1e-6,
                                                                   show_text_editor=True, show_save_button=True))


@dataclass
class ImageCentralAxes:
    x: bool = False
    y: bool = False


@dataclass
class Image(ViewerElement):
    wcs_transform: bool = False
    rotate: float | str | None = None
    color_bar: ColorBar = field(default_factory=ColorBar)
    central_axes: ImageCentralAxes = field(default_factory=ImageCentralAxes)
    central_crosshair: bool = False


@dataclass
class Plot1D(ViewerElement):
    plots: dict[str, LinePlot] = field(default_factory=dict)


@dataclass
class DataWidgets(Params):
    images: dict[str, Image] = field(default_factory=dict)
    plots: dict[str, Plot1D] = field(default_factory=dict)



from specvizitor.io.data_dir import _get_id_from_filename, get_ids_from_dir


def test_get_id_from_filename():
    obj_id = _get_id_from_filename("stacked_2D_F300M_1.fits", r"\d+")
    assert obj_id == "300"

    obj_id = _get_id_from_filename("stacked_2D_F300M_1.fits", r"(?<![0-9Ff])\d+")
    assert obj_id == "2"

    obj_id = _get_id_from_filename("stacked_2D_F300M_1.fits", r"(?<![0-9Ff])\d+(?![0-9Dd])")
    assert obj_id == "1"

    obj_id = _get_id_from_filename("stacked_2D_F300M_1_234.fits", r"(?<![0-9Ff])\d+(?![0-9Dd])")
    assert obj_id == "234"


def test_get_ids_from_dir(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)

    (tmp_path / "123.fits").touch()
    wfss_dir = tmp_path / "wfss"
    wfss_dir.mkdir()
    (wfss_dir / "456.fits").touch()

    ids = get_ids_from_dir(tmp_path)
    assert list(ids) == [123]

    ids = get_ids_from_dir(tmp_path, recursive=True)
    assert list(ids) == [123, 456]



from qtpy import QtWidgets, QtCore

from functools import partial


class TableColumns(QtWidgets.QDialog):
    visible_columns_updated = QtCore.Signal(list)

    def __init__(self, colnames: list[str], visible_columns: list[str], parent=None):
        self.colnames = colnames
        self.visible_columns = visible_columns

        self._check_all: QtWidgets.QPushButton | None = None
        self._uncheck_all: QtWidgets.QPushButton | None = None

        self._table: QtWidgets.QTableWidget | None = None
        self._checkboxes: list[QtWidgets.QCheckBox] | None = None
        self._button_box: QtWidgets.QDialogButtonBox | None = None

        super().__init__(parent)
        self.setWindowTitle("Table Columns")

        self.init_ui()
        self.set_layout()
        self.populate()

    def _create_checkboxes(self):
        checkboxes = []
        for i, name in enumerate(self.colnames):
            checkbox = QtWidgets.QCheckBox(self)
            checkbox.setChecked(name in self.visible_columns)
            checkboxes.append(checkbox)

        self._checkboxes = checkboxes

    def _add_table_items(self):
        for i, name in enumerate(self.colnames):
            checkbox_container = QtWidgets.QWidget(self)
            sub_layout = QtWidgets.QHBoxLayout()
            sub_layout.setAlignment(QtCore.Qt.AlignCenter)
            sub_layout.addWidget(self._checkboxes[i])
            checkbox_container.setLayout(sub_layout)

            self._table.setCellWidget(i, 0, checkbox_container)
            self._table.setItem(i, 1, QtWidgets.QTableWidgetItem(name))

    def init_ui(self):
        self._check_all = QtWidgets.QPushButton('Check All', self)
        self._uncheck_all = QtWidgets.QPushButton('Uncheck All', self)

        self._table = QtWidgets.QTableWidget(len(self.colnames), 2, self)
        self._table.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)
        self._table.setHorizontalHeaderLabels(('Visible', 'Name'))
        self._table.horizontalHeader().resizeSection(0, 100)
        self._table.horizontalHeader().setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
        self._table.setSizeAdjustPolicy(QtWidgets.QAbstractScrollArea.AdjustToContents)

        self._create_checkboxes()
        self._add_table_items()

        # add OK/Cancel buttons
        self._button_box = QtWidgets.QDialogButtonBox(
            QtWidgets.QDialogButtonBox.Ok | QtWidgets.QDialogButtonBox.Cancel, self)

        self._check_all.clicked.connect(partial(self._all_action, True))
        self._uncheck_all.clicked.connect(partial(self._all_action, False))

        self._button_box.accepted.connect(self.accept)
        self._button_box.rejected.connect(self.reject)

    def set_layout(self):
        self.setLayout(QtWidgets.QVBoxLayout())

    def populate(self):
        sub_layout = QtWidgets.QHBoxLayout()
        sub_layout.addWidget(self._check_all)
        sub_layout.addWidget(self._uncheck_all)
        self.layout().addLayout(sub_layout)

        self.layout().addWidget(self._table)
        self.layout().addWidget(self._button_box)

    def accept(self):
        self.visible_columns = [cname for i, cname in enumerate(self.colnames) if self._checkboxes[i].isChecked()]
        self.visible_columns_updated.emit(self.visible_columns)

        super().accept()

    @QtCore.Slot()
    def _all_action(self, state: bool):
        for c in self._checkboxes:
            c.setChecked(state)



import qtpy.compat
from qtpy import QtWidgets, QtCore, QtGui
import numpy as np

from dataclasses import asdict
from importlib import metadata
import logging
import pathlib
import time

from ..appdata import AppData
from ..config.appearance import setup_appearance
from ..config import Config, Cache, DataWidgets, SpectralLineData
from ..io.catalog import Catalog
from ..io.inspection_data import InspectionData
from ..plugins.plugin_core import PluginCore
from ..utils.params import save_yaml

from .DataViewer import DataViewer
from .NewFile import NewFile
from .ObjectInfo import ObjectInfo
from .QuickSearch import QuickSearch
from .InspectionResults import InspectionResults
from .InspectionFieldEditor import InspectionFieldEditor
from .Subsets import Subsets
from .Settings import Settings
from .ToolBar import ToolBar
from ..utils.widgets import AbstractWidget

logger = logging.getLogger(__name__)


class MainWindow(QtWidgets.QMainWindow):
    EXIT_CODE_REBOOT = -42
    project_loaded = QtCore.Signal(InspectionData)
    object_selected = QtCore.Signal(int, InspectionData, object)
    loading_aborted = QtCore.Signal()
    data_requested = QtCore.Signal()
    save_action_invoked = QtCore.Signal()
    delete_action_invoked = QtCore.Signal()
    close_action_invoked = QtCore.Signal()

    catalogue_changed = QtCore.Signal(object)
    inspection_fields_changed = QtCore.Signal(int, InspectionData)
    data_source_changed = QtCore.Signal()
    spectral_lines_changed = QtCore.Signal()
    visible_columns_updated = QtCore.Signal(list)
    dock_layout_updated = QtCore.Signal(dict)
    viewer_configuration_updated = QtCore.Signal(DataWidgets)

    starred_state_updated = QtCore.Signal(bool, bool)
    screenshot_path_selected = QtCore.Signal(str)

    subset_loaded = QtCore.Signal(str, object)
    is_subset_inspection_paused = QtCore.Signal(bool)
    subset_inspection_stopped = QtCore.Signal()

    is_zen_mode_activated = QtCore.Signal(bool)

    def __init__(self,
                 config: Config | None = None,
                 cache: Cache | None = None,
                 widget_cfg: DataWidgets | None = None,
                 spectral_lines: SpectralLineData | None = None,
                 plugins: list[PluginCore] | None = None,
                 parent=None):

        super().__init__(parent)

        self.rd = AppData()

        self._config = config if config else Config()
        self._cache = cache if cache else Cache()
        self._widget_cfg = widget_cfg if widget_cfg else DataWidgets()
        self._spectral_lines = spectral_lines if spectral_lines else SpectralLineData()
        self._plugins: list[PluginCore] = plugins if plugins is not None else []

        self._object_loaded: bool = False
        self._t_load_object_start = None
        self._restart_requested: bool = False
        self._zen_mode_activated: bool = False
        self._was_maximized: bool = False

        self._subset_cat: Catalog | None = None
        self._subset_inspection_paused: bool = False

        self.setFocusPolicy(QtCore.Qt.StrongFocus)

        self._update_window_title()  # set the title of the main window
        # self.setWindowIcon(QtGui.QIcon('logo2_2.png'))

        # add a status bar
        # self.statusBar().showMessage("Message in the statusbar")

        QtWidgets.QShortcut('Ctrl+S', self, self.save_action_invoked.emit)
        QtWidgets.QShortcut('Ctrl+D', self, self.delete_action_invoked.emit)
        QtWidgets.QShortcut('Esc', self, lambda: self._exit_fullscreen() if self.isFullScreen() else None)

        self._data_viewer: DataViewer | None = None
        self._commands_bar: ToolBar | None = None
        self._quick_search: QuickSearch | None = None
        self._object_info: ObjectInfo | None = None
        self._inspection_res: InspectionResults | None = None
        self._subsets: Subsets | None = None

        self._quick_search_dock: QtWidgets.QDockWidget | None = None
        self._object_info_dock: QtWidgets.QDockWidget | None = None
        self._inspection_res_dock: QtWidgets.QDockWidget | None = None
        self._subsets_dock: QtWidgets.QDockWidget | None = None

        self.init_ui()
        self.populate()
        self.connect()

        self.restore_window_state()

        # restore the dock state (= the layout of the data viewer) from cache
        if self._cache.dock_layout:
            self.dock_layout_updated.emit(self._cache.dock_layout)

        # load the catalogue to the memory
        if self._config.catalogue.filename:
            self.load_catalogue()

        # update the list of catalogue columns visible in the object info widget
        # write "is not None" explicitly in case visible_columns == []
        if self.rd.cat and self._cache.visible_columns is not None:
            self.visible_columns_updated.emit(self._cache.visible_columns)

        # load the subset to the memory
        if self._cache.last_subset_file:
            self.load_subset(reset_index=False)

        # read cache and try to load the last opened project
        if self._cache.last_inspection_file:
            self.open_file(self._cache.last_inspection_file, self._cache.last_object_index)

    def init_ui(self):
        # create a central widget
        self._data_viewer = DataViewer(self._config.data_viewer, self._config.data, self._widget_cfg,
                                       self._config.appearance, spectral_lines=self._spectral_lines,
                                       plugins=self._plugins, parent=self)

        # create a toolbar
        self._commands_bar = ToolBar(self._config.appearance, parent=self)
        self._commands_bar.setObjectName('Toolbar')
        self._commands_bar.setEnabled(True)

        # create a quick search widget
        self._quick_search = QuickSearch(parent=self)
        self._quick_search_dock = QtWidgets.QDockWidget('Quick Search', self)
        self._quick_search_dock.setObjectName('Quick Search')
        self._quick_search_dock.setWidget(self._quick_search)

        # create a widget displaying information about the object
        self._object_info = ObjectInfo(parent=self)
        self._object_info_dock = QtWidgets.QDockWidget('Object Information', self)
        self._object_info_dock.setObjectName('Object Information')
        self._object_info_dock.setWidget(self._object_info)

        # create a widget for writing comments
        self._inspection_res = InspectionResults(cfg=self._config.inspection_results, parent=self)
        self._inspection_res_dock = QtWidgets.QDockWidget('Inspection Results', self)
        self._inspection_res_dock.setObjectName('Inspection Results')
        self._inspection_res_dock.setWidget(self._inspection_res)

        # create a widget for inspecting a subset of objects
        self._subsets = Subsets(parent=self)
        self._subsets_dock = QtWidgets.QDockWidget('Subsets', self)
        self._subsets_dock.setObjectName('Subsets')
        self._subsets_dock.setWidget(self._subsets)

        self._init_menu()

    def _init_menu(self):
        self._menu = self.menuBar()

        self._file = self._menu.addMenu("&File")

        self._new_file = QtWidgets.QAction("&New...")
        self._new_file.triggered.connect(self._new_file_action)
        self._new_file.setShortcut(QtGui.QKeySequence('Ctrl+N'))
        self._file.addAction(self._new_file)

        self._open_file = QtWidgets.QAction("&Open...")
        self._open_file.triggered.connect(self._open_file_action)
        self._open_file.setShortcut(QtGui.QKeySequence('Ctrl+O'))
        self._file.addAction(self._open_file)

        self._file.addSeparator()

        self._export = QtWidgets.QAction("&Export FITS Table...")
        self._export.triggered.connect(self._export_action)
        self._export.setEnabled(False)
        self._file.addAction(self._export)

        self._file.addSeparator()

        self._quit = QtWidgets.QAction("&Quit")
        self._quit.triggered.connect(self.close)
        self._quit.setShortcut(QtGui.QKeySequence('Ctrl+Q'))
        self._file.addAction(self._quit)

        self._view = self._menu.addMenu("&View")

        self._reset_view = QtWidgets.QAction("Reset View")
        self._reset_view.setEnabled(False)
        self._reset_view.triggered.connect(self._data_viewer.reset_view)
        self._reset_view.setShortcut('R')
        self._view.addAction(self._reset_view)

        self._reset_dock_layout = QtWidgets.QAction("Reset Layout")
        self._reset_dock_layout.setEnabled(False)
        self._reset_dock_layout.triggered.connect(self._data_viewer.reset_dock_layout)
        self._view.addAction(self._reset_dock_layout)

        self._view.addSeparator()

        self._zen = QtWidgets.QAction("Enter Zen Mode")
        self._zen.triggered.connect(self._zen_mode_action)
        self._zen.setShortcut('H')
        self._auxiliary_docks = (self._quick_search_dock, self._object_info_dock, self._subsets_dock)
        self._view.addAction(self._zen)

        self._view.addSeparator()

        self._fullscreen = QtWidgets.QAction("Fullscreen")
        self._fullscreen.triggered.connect(lambda:
                                           self._exit_fullscreen() if self.isFullScreen() else self._enter_fullscreen())
        self._fullscreen.setShortcut('F')
        self._view.addAction(self._fullscreen)

        self._widgets = self._menu.addMenu("&Widgets")

        self._backup_viewer_config = QtWidgets.QAction("Backup...")
        self._backup_viewer_config.triggered.connect(self._backup_viewer_config_action)
        self._widgets.addAction(self._backup_viewer_config)

        self._restore_viewer_config = QtWidgets.QAction("Restore...")
        self._restore_viewer_config.triggered.connect(self._restore_viewer_config_action)
        self._widgets.addAction(self._restore_viewer_config)

        self._tools = self._menu.addMenu("&Tools")

        self._inspect_subset = QtWidgets.QAction("Inspect Subset...")
        self._inspect_subset.triggered.connect(self._inspect_subset_action)
        self._tools.addAction(self._inspect_subset)

        self._screenshot = QtWidgets.QAction("Take Screenshot...")
        self._screenshot.triggered.connect(self._screenshot_action)
        self._tools.addAction(self._screenshot)

        self._tools.addSeparator()

        self._settings = QtWidgets.QAction("Se&ttings...")
        self._settings.triggered.connect(self._settings_action)
        self._tools.addAction(self._settings)

        self._help = self._menu.addMenu("&Help")
        self._about = QtWidgets.QAction("&About...")
        self._about.triggered.connect(self._about_action)
        self._help.addAction(self._about)

    def connect(self):
        # connect the main window to the child widgets
        for w in (self._data_viewer, self._commands_bar, self._quick_search, self._object_info, self._inspection_res,
                  self._subsets):
            self.project_loaded.connect(w.load_project)

        for w in (self._data_viewer, self._object_info, self._inspection_res):
            self.data_requested.connect(w.collect)

        self.loading_aborted.connect(self._data_viewer.abort_loading)

        self.catalogue_changed.connect(self._object_info.update_table_items)
        self.inspection_fields_changed.connect(self._inspection_res.update_inspection_fields)
        self.data_source_changed.connect(self._data_viewer.open_images)
        self.spectral_lines_changed.connect(self._data_viewer.spectral_lines_changed.emit)
        self.visible_columns_updated.connect(self._object_info.update_visible_columns)
        self.dock_layout_updated.connect(self._data_viewer.restore_dock_layout)
        self.viewer_configuration_updated.connect(self._data_viewer.update_viewer_configuration)

        self.starred_state_updated.connect(self._commands_bar.update_star_button_icon)
        self.screenshot_path_selected.connect(self._data_viewer.take_screenshot)

        self.subset_loaded.connect(self._subsets.load_subset)
        self.is_subset_inspection_paused.connect(self._subsets.pause_subset_inspection)
        self.subset_inspection_stopped.connect(self._subsets.stop_subset_inspection)

        self.save_action_invoked.connect(self._data_viewer.request_redshift)
        self.delete_action_invoked.connect(self._inspection_res.clear_redshift_value)
        self.close_action_invoked.connect(self._data_viewer.free_resources)

        self.is_zen_mode_activated.connect(self._data_viewer.enter_zen_mode)

        # connect the child widgets to the main window
        self._quick_search.id_selected.connect(self.load_by_id)
        self._quick_search.index_selected.connect(self.load_by_index)

        self._subsets.inspect_button_clicked.connect(self._inspect_subset_action)
        self._subsets.pause_inspecting_button_clicked.connect(self._pause_subset_inspection_action)
        self._subsets.stop_inspecting_button_clicked.connect(self._stop_subset_inspection_action)

        self._commands_bar.navigation_button_clicked.connect(self.switch_object)
        self._commands_bar.star_button_clicked.connect(self.update_starred_state)
        self._commands_bar.screenshot_button_clicked.connect(self._screenshot_action)
        self._commands_bar.settings_button_clicked.connect(self._settings_action)

        self._inspection_res.edit_button_clicked.connect(self._edit_inspection_file_action)

        self._data_viewer.object_loaded.connect(self.finalize_loading)

        self._data_viewer.data_collected.connect(self.save_viewer_data)
        self._object_info.data_collected.connect(self.save_obj_info_data)
        self._inspection_res.data_collected.connect(self.save_review_data)

        # connect the child widgets between each other
        self._data_viewer.redshift_obtained.connect(self._inspection_res.set_redshift_value)
        self._commands_bar.reset_view_button_clicked.connect(self._data_viewer.reset_view)
        self._commands_bar.reset_layout_button_clicked.connect(self._data_viewer.reset_dock_layout)

    def populate(self):
        self.setCentralWidget(self._data_viewer)

        self.addToolBar(QtCore.Qt.TopToolBarArea, self._commands_bar)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self._quick_search_dock)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self._object_info_dock)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self._inspection_res_dock)
        self.addDockWidget(QtCore.Qt.RightDockWidgetArea, self._subsets_dock)

    def load_catalogue(self):
        cat = Catalog.read(self._config.catalogue.filename, translate=self._config.catalogue.translate)
        if cat is None:
            self._config.catalogue.filename = None
            self._config.save()

        self.update_catalogue(cat)

    def restore_window_state(self):
        settings = QtCore.QSettings()
        if settings.value("geometry") is None or settings.value("windowState") is None:
            self.showMaximized()
            self.setFocus()
        else:
            self.restoreGeometry(settings.value("geometry"))
            self.restoreState(settings.value("windowState"))

    def _new_file_action(self):
        """ Create a new inspection file via the NewFile dialog.
        """
        dialog = NewFile(cfg=self._config, parent=self)
        dialog.catalogue_changed.connect(self.update_catalogue)
        dialog.output_path_selected.connect(self.update_output_path)
        if dialog.exec():
            self.rd.create(flags=self._config.inspection_results.default_flags)
            if self.rd.review is not None:
                self.load_project()
            else:
                self.update_catalogue(None)
                self.update_output_path(None)

    def _open_file_action(self):
        """ Open an existing inspection file via QFileDialog.
        """
        path = qtpy.compat.getopenfilename(self, caption='Open Inspection File', filters='CSV Files (*.csv)')[0]
        if path:
            self.open_file(path)

    def open_file(self, path: str, cached_index: int | None = None):
        """ Load inspection data from an existing inspection file.
        @param path: path to the inspection file
        @param cached_index:
        """
        path = pathlib.Path(path)
        if path.exists():
            self.update_output_path(path)
            self.rd.read()
            if self.rd.cat is None:
                self.update_catalogue(None)  # in case the catalogue hasn't been initialized before
            self.load_project(cached_index)
        else:
            logger.warning('Inspection file not found (path: {})'.format(path))

    def load_project(self, j: int | None = None):
        """ Update the state of the main window and activate the central widget after loading inspection data.
        """
        for w in (self._export, self._reset_view, self._reset_dock_layout):
            w.setEnabled(True)

        self.project_loaded.emit(self.rd.review)

        if j is None or (j < 0 or j >= self.rd.review.n_objects):
            j = 0

        self._object_loaded = False
        self.load_object(j)

    @QtCore.Slot(int)
    def load_object(self, j: int):
        """ Load a new object to the central widget.
        @param j: the index of the object to display
        """
        if self._object_loaded:
            self.data_requested.emit()
        else:
            self.loading_aborted.emit()

        self._object_loaded = False
        self._t_load_object_start = time.perf_counter()

        self.rd.j = j

        self._load_object(self._data_viewer)
        self._update_window_title()

    def _get_cat_entry(self) -> Catalog | None:
        if self.rd.j is None:
            return None

        obj_id = self.rd.review.get_id(self.rd.j, full=True)
        return self.rd.cat.get_cat_entry(obj_id)

    def _load_object(self, *widgets):
        for w in widgets:
            self.object_selected.connect(w.load_object)
        self.object_selected.emit(self.rd.j, self.rd.review, self._get_cat_entry())
        for w in widgets:
            self.object_selected.disconnect(w.load_object)

    @QtCore.Slot()
    def finalize_loading(self):
        self._load_object(self._commands_bar, self._inspection_res,self._subsets, self._object_info)
        self._object_loaded = True

        self._cache.last_object_index = self.rd.j
        self._cache.save()

        logger.info(f"Object loaded (ID: {self.rd.review.get_id(self.rd.j)}, loading time: {time.perf_counter()-self._t_load_object_start:.3f} s)")

    @QtCore.Slot(str, bool)
    def switch_object(self, command: str, switch_to_starred: bool):
        if command not in ('next', 'previous'):
            logger.error(f'Unknown command: {command}')
            return

        if switch_to_starred and not self.rd.review.has_data('starred'):
            logger.error('No starred objects found')
            return

        j_upd = self._update_index(self.rd.j, command)
        if self._subset_cat and not self._subset_inspection_paused:
            while True:
                subset_entry = self._subset_cat.get_cat_entry(self.rd.review.get_id(j_upd, full=True),
                                                              ignore_missing=True)
                if subset_entry is not None:
                    break

                j_upd = self._update_index(j_upd, command)
                if j_upd == self.rd.j:  # in case no other ID from the subset is found
                    logger.warning('No other IDs found in the subset')
                    break

        elif switch_to_starred:
            while not self.rd.review.get_value(j_upd, 'starred'):
                j_upd = self._update_index(j_upd, command)

        self.load_object(j_upd)

    def _update_index(self, obj_index: int, command: str) -> int:
        j_upd = obj_index

        if command == 'next':
            j_upd += 1
        elif command == 'previous':
            j_upd -= 1

        j_upd = j_upd % self.rd.review.n_objects
        return j_upd

    @QtCore.Slot(str)
    def load_by_id(self, obj_id: str):
        if self.rd.review.validate_id(obj_id):
            self.setFocus()
            self.load_object(self.rd.review.get_id_loc(obj_id))

    @QtCore.Slot(int)
    def load_by_index(self, index: int):
        if self.rd.review.validate_index(index):
            self.setFocus()
            self.load_object(index - 1)

    def _reload(self):
        if self.rd.j is not None:
            self.load_object(self.rd.j)

    def _update_window_title(self):
        title = ''
        if self.rd.output_path is not None:
            title += f'{self.rd.output_path.name} – '
        if self.rd.j is not None:
            obj_id = self.rd.review.get_id(self.rd.j, full=True)
            obj_id = obj_id if isinstance(obj_id, tuple) else (obj_id,)
            id_str = ', '.join(map(str, obj_id))
            title += f"ID {id_str} [#{self.rd.j + 1}/{self.rd.review.n_objects}] – "
        title += 'Specvizitor'
        self.setWindowTitle(title)

    def _export_action(self):
        path = qtpy.compat.getsavefilename(self, caption='Export To FITS',
                                           basedir=str(self.rd.output_path.with_suffix('.fits')),
                                           filters='FITS Files (*.fits)')[0]

        if path:
            self.rd.review.write(path, 'fits')

    def _restore_viewer_config_action(self):
        path = qtpy.compat.getopenfilename(self, caption='Open Viewer Configuration',
                                           filters='YAML Files (*.yml)')[0]
        if path:
            new_viewer_cfg = self._widget_cfg.replace_params(pathlib.Path(path))
            if new_viewer_cfg is None:
                logger.error('Failed to restore the viewer configuration')
            else:
                self._widget_cfg = new_viewer_cfg
                self._widget_cfg.save()

                self.viewer_configuration_updated.emit(self._widget_cfg)

                self._reload()

                logger.info('Viewer configuration restored')

    def _backup_viewer_config_action(self):
        path = qtpy.compat.getsavefilename(self, caption='Save Viewer Configuration',
                                           basedir=str(pathlib.Path() / 'data_widgets.yml'),
                                           filters='YAML Files (*.yml)')[0]

        if path:
            save_yaml(path, asdict(self._widget_cfg))
            logger.info('Viewer configuration saved')

    def _zen_mode_action(self):
        self._zen_mode_activated = not self._zen_mode_activated
        self._zen.setText("Exit Zen Mode" if self._zen_mode_activated else "Enter Zen Mode")
        for w in self._auxiliary_docks:
            w.setVisible(not self._zen_mode_activated)
        self.is_zen_mode_activated.emit(self._zen_mode_activated)

    def _enter_fullscreen(self):
        self._was_maximized = True if self.isMaximized() else False
        self.showFullScreen()

    def _exit_fullscreen(self):
        self.showNormal()
        if self._was_maximized:
            self.showMaximized()

    def _edit_inspection_file_action(self):
        dialog = InspectionFieldEditor(review=self.rd.review, parent=self)
        dialog.inspection_fields_updated.connect(self.update_inspection_fields)
        if dialog.exec():
            self.inspection_fields_changed.emit(self.rd.j, self.rd.review)

    def _inspect_subset_action(self):
        path = qtpy.compat.getopenfilename(self, caption='Open Subset')[0]
        if path:
            self._cache.last_subset_file = path
            self._cache.save()
            self.load_subset()

    def load_subset(self, reset_index=True):
        subset_path = self._cache.last_subset_file
        subset = Catalog.read(subset_path, translate=self._config.catalogue.translate)
        if subset:
            subset.add_column(np.arange(len(subset)), name='__index__', index=0)

            self._subset_cat = subset
            self.subset_loaded.emit(subset_path, subset)

            if reset_index:
                self.load_by_id(self._subset_cat.get_col('id')[0])
        else:
            self._cache.last_subset_file = None
            self._cache.save()

    def _pause_subset_inspection_action(self):
        self._subset_inspection_paused = not self._subset_inspection_paused
        self.is_subset_inspection_paused.emit(self._subset_inspection_paused)

    def _stop_subset_inspection_action(self):
        self._subset_cat = None
        self._subset_inspection_paused = False

        self._cache.last_subset_file = None
        self._cache.save()

        self.subset_inspection_stopped.emit()

    def _screenshot_action(self):
        default_filename = f'{self.rd.output_path.stem.replace(" ", "_")}_ID{self.rd.review.get_id(self.rd.j)}.png'
        path, extension = qtpy.compat.getsavefilename(self, caption='Save/Save As',
                                                      basedir=str(pathlib.Path().resolve() / default_filename),
                                                      filters='Images (*.png)')
        if path:
            self.screenshot_path_selected.emit(path)

    def _settings_action(self):
        self._restart_requested = False

        dialog = Settings(self.rd.cat, self._config, self._spectral_lines, parent=self)

        dialog.appearance_changed.connect(self.update_appearance)
        dialog.catalogue_changed.connect(self.update_catalogue)
        dialog.spectral_lines_changed.connect(self.spectral_lines_changed.emit)
        dialog.data_source_changed.connect(self.data_source_changed.emit)

        dialog.restart_requested.connect(self.restart)

        if dialog.exec():
            if not self._restart_requested:
                self._reload()

    @QtCore.Slot()
    def update_appearance(self):
        setup_appearance(self._config.appearance, update_theme=False)

    @QtCore.Slot(object)
    def update_catalogue(self, cat: Catalog | None):
        if cat is None and self.rd.review is not None:
            self.rd.cat = Catalog.create(self.rd.review.ids_full)
        else:
            self.rd.cat = cat
        self.catalogue_changed.emit(self.rd.cat)

    @QtCore.Slot(pathlib.Path)
    def update_output_path(self, path: pathlib.Path | None):
        if self._object_loaded:
            self.data_requested.emit()

        self.rd.output_path = path
        self._cache.last_inspection_file = None if path is None else str(path)
        self._cache.save()
        self._update_window_title()

    @QtCore.Slot(list, list, bool)
    def update_inspection_fields(self, fields: list[tuple[str, str]], is_deleted: list[bool], set_as_default: bool):
        old_columns = self.rd.review.user_defined_columns
        for i, old_name in enumerate(old_columns):
            new_name = fields[i][0]
            if is_deleted[i]:
                self.rd.review.delete_column(column_name=old_name)
            elif old_name != new_name:
                self.rd.review.rename_column(old_name=old_name, new_name=new_name)

        for field in fields[len(old_columns):]:
            new_name, new_type = field[0], field[1]
            if new_type == 'boolean':
                self.rd.review.add_flag_column(column_name=new_name)

        if set_as_default:
            default_flags = self.rd.review.flag_columns
            self._config.inspection_results.default_flags = default_flags
            self._config.save()

    def _about_action(self):
        package_name: str = __package__.split('.')[0]

        msg = []
        msg.append(f"{package_name.capitalize()} v{metadata.version(package_name)}")
        msg.append(f"Copyright (c) 2025, Ivan Kramarenko, Josephine Kerutt")
        msg.append(f"\nProject's homepage: {metadata.metadata(package_name).get('Home-page')}")

        QtWidgets.QMessageBox.about(self, f"About {package_name.capitalize()}", '\n'.join(msg))

    def closeEvent(self, a0):
        if self.rd.j is not None:
            self.data_requested.emit()

        self.close_action_invoked.emit()

        # save the state and geometry of the main window
        if self._zen_mode_activated:
            self._zen_mode_action()
        settings = QtCore.QSettings()
        settings.setValue('geometry', self.saveGeometry())
        settings.setValue('windowState', self.saveState())

        super().closeEvent(a0)

    @QtCore.Slot()
    def update_starred_state(self):
        starred = not self.rd.review.get_value(self.rd.j, 'starred')
        self.rd.review.update_value(self.rd.j, 'starred', starred)

        self.starred_state_updated.emit(starred, self.rd.review.has_data('starred'))

    @QtCore.Slot(dict)
    def save_viewer_data(self, layout: dict):
        self._cache.dock_layout = layout
        self._cache.save()

    @QtCore.Slot(list)
    def save_obj_info_data(self, visible_columns: list[str]):
        self._cache.visible_columns = visible_columns
        self._cache.save()

    @QtCore.Slot(float, str, dict)
    def save_review_data(self, redshift: float, comments: str, checkboxes: dict[str, bool]):
        self.rd.review.update_value(self.rd.j, 'z_sviz', redshift)
        self.rd.review.update_value(self.rd.j, 'comment', comments)
        for cname, is_checked in checkboxes.items():
            self.rd.review.update_value(self.rd.j, cname, is_checked)
        self.rd.save()

    @QtCore.Slot()
    def restart(self):
        self._restart_requested = True
        self.close()
        QtWidgets.QApplication.exit(MainWindow.EXIT_CODE_REBOOT)



from qtpy import QtWidgets, QtCore

import logging

from ..utils.widgets import AbstractWidget


logger = logging.getLogger(__name__)


class QuickSearch(AbstractWidget):
    id_selected = QtCore.Signal(str)
    index_selected = QtCore.Signal(int)

    def __init__(self, parent=None):
        self._go_to_id_button: QtWidgets.QPushButton | None = None
        self._id_field: QtWidgets.QLineEdit | None = None
        self._go_to_index_button: QtWidgets.QPushButton | None = None
        self._index_field: QtWidgets.QLineEdit | None = None

        super().__init__(parent=parent)
        self.setEnabled(False)
        self.setSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Maximum)

    def init_ui(self):
        # create a `Go to ID` button
        self._go_to_id_button = QtWidgets.QPushButton(self)
        self._go_to_id_button.setText('Go to ID')
        self._go_to_id_button.setFixedWidth(90)

        self._id_field = QtWidgets.QLineEdit(self)
        self._id_field.setMinimumWidth(60)

        # create a `Go to index` button
        self._go_to_index_button = QtWidgets.QPushButton(self)
        self._go_to_index_button.setText('Go to #')
        self._go_to_index_button.setFixedWidth(90)

        self._index_field = QtWidgets.QLineEdit(self)
        self._index_field.setMinimumWidth(60)

        self._go_to_id_button.clicked.connect(self.go_to_id)
        self._id_field.returnPressed.connect(self.go_to_id)
        self._go_to_index_button.clicked.connect(self.go_to_index)
        self._index_field.returnPressed.connect(self.go_to_index)

    def set_layout(self):
        self.setLayout(QtWidgets.QGridLayout())

    def populate(self):
        self.layout().addWidget(self._go_to_id_button, 1, 1, 1, 1)
        self.layout().addWidget(self._id_field, 1, 2, 1, 1)
        self.layout().addWidget(self._go_to_index_button, 2, 1, 1, 1)
        self.layout().addWidget(self._index_field, 2, 2, 1, 1)

    @QtCore.Slot()
    def load_project(self):
        self.setEnabled(True)

    def go_to_id(self):
        # don't validate the ID here because it could be either int OR str
        self.id_selected.emit(self._id_field.text())
        self._id_field.clear()

    def go_to_index(self):
        text = self._index_field.text()
        self._index_field.clear()

        # validate the index (must be int)
        try:
            index = int(text)
        except ValueError:
            logger.error(f'Invalid index: {text}')
            return False

        self.index_selected.emit(index)



from astropy.io import fits
from astropy.table import Table
import astropy.units as u
from astropy.utils.exceptions import AstropyWarning
from astropy.wcs import WCS
import numpy as np
from PIL import Image, ImageOps
import rasterio

import abc
from collections import OrderedDict
import logging
import pathlib
from string import Formatter
from typing import Any
import warnings

from .catalog import Catalog
from ..utils.widgets import FileBrowser


__all__ = [
    "ViewerData",
    "DataPath",
    "LocalPath",
    "URLPath",
    "add_unit_aliases",
    "data_browser"
]

Image.MAX_IMAGE_PIXELS = None  # ignore warnings when loading large images
logger = logging.getLogger(__name__)


class BaseLoader(abc.ABC):
    name: str
    extensions: tuple[str, ...] = ()

    def __init__(self):
        self._dataset = None
        self._meta: dict | None = None

        self._last_kwargs: dict | None = None
        self.last_data: tuple[Any, Any] | None = None

    def open(self, filename: str, **kwargs):
        self._open(filename, **kwargs)
        self._last_kwargs = kwargs

    @abc.abstractmethod
    def _open(self, filename: str, **kwargs):
        pass

    def reopen(self, filename: str):
        if self._last_kwargs is None:
            raise RuntimeError("Failed to re-open the file: file was never opened")
        self.open(filename, **self._last_kwargs)

    def load(self, **kwargs) -> tuple[Any, Any]:
        self.last_data = self._load(**kwargs)
        return self.last_data

    def _load(self, **kwargs):
        return self._dataset, self._meta

    def close(self):
        self._dataset.close()

    @classmethod
    def validate_extension(cls, filename: str | pathlib.Path) -> bool:
        return any(str(filename).endswith(s) for s in cls.extensions)

    @staticmethod
    def get_cutout_params(arr_shape, x0=None, y0=None, cutout_size=100, **kwargs) -> tuple[tuple[float, float, float, float], float]:
        if x0 is None:
            x0 = arr_shape[1] // 2
        if y0 is None:
            y0 = arr_shape[0] // 2

        x0, y0 = int(x0), int(y0)

        x1, x2 = x0 - cutout_size, x0 + cutout_size
        y1, y2 = y0 - cutout_size, y0 + cutout_size

        return (x1, x2, y1, y2), cutout_size


class GenericFITSLoader(BaseLoader):
    name = 'generic_fits'
    extensions = ('.fits', '.fits.gz')

    def _open(self, filename: str, **kwargs):
        self._dataset = fits.open(filename, **kwargs)

    def _load(self, extname: str = None, extver: str = None, extver_index: int = None, create_cutout=False,
              create_wcs=False, **kwargs):
        hdul = self._dataset

        if extname is not None and extver is not None:
            index = (extname, extver)
        elif extname is not None and extver_index is not None:
            extname_match_indices = [i for i, hdu in enumerate(hdul) if hdu.header.get('EXTNAME') == extname]
            try:
                index = extname_match_indices[extver_index]
            except IndexError:
                IndexError(f"EXTVER `{extver_index}` out of range")
        elif extname is not None:
            index = extname
        elif len(hdul) > 1:
            index = 1
        else:
            index = 0

        try:
            hdu = hdul[index]
        except KeyError:
            KeyError(f"Extension `{index}` not found")

        meta = hdu.header
        if meta.get('XTENSION') and meta['XTENSION'] in ('TABLE', 'BINTABLE'):
            data = Table.read(hdu)
        else:
            data = hdu.data

        if create_cutout:
            coords, _ = self.get_cutout_params(data.shape, **kwargs)
            data = self._create_cutout(data, *coords)

        if create_wcs:
            meta.wcs = self._create_wcs(meta)

        return data, meta

    @staticmethod
    def _create_cutout(data, x1, x2, y1, y2):
        return data[y1:y2, x1:x2]

    @staticmethod
    def _create_wcs(meta):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore', AstropyWarning)
            wcs = WCS(meta)
        return wcs


class PILLoader(BaseLoader):
    name = 'pil'

    def _open(self, filename: pathlib.Path, **kwargs):
        image = Image.open(filename, **kwargs)
        image = ImageOps.flip(image)
        self._dataset, self._meta = np.array(image), image.info


class RasterIOLoader(BaseLoader):
    name = 'rasterio'
    extensions = ('.tif', '.tiff')

    def _open(self, filename: pathlib.Path, **kwargs):
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=rasterio.errors.NotGeoreferencedWarning)
            self._dataset = rasterio.open(filename, **kwargs)

    def _load(self, create_cutout=False, **kwargs):
        if create_cutout:
            (x1, x2, y1, y2), cutout_size = self.get_cutout_params(self._dataset.shape, **kwargs)
            data = self._create_cutout(x1, y2, cutout_size)
        else:
            data = self._dataset.read()

        return data, self._meta

    def _create_cutout(self, x1, y2, cutout_size):
        # wow, this actually works!
        data = self._dataset.read(window=rasterio.windows.Window(x1, self._dataset.height - y2,
                                                                 2 * cutout_size, 2 * cutout_size))
        data = np.moveaxis(data, 0, -1)
        data = np.flip(data, 0)
        return data


class ViewerData:
    def __init__(self):
        self._loaders: dict[str, BaseLoader] = {}
        self._loader_constructors: OrderedDict[str, type(BaseLoader)] = OrderedDict(
            [(loader.name, loader) for loader in (GenericFITSLoader, RasterIOLoader, PILLoader)]
        )

    @property
    def loader_names(self):
        return ('auto',) + tuple(loader.name for loader in self._loader_constructors.values())

    def _get_loader(self, filename: str) -> str:
        for ln, loader in self._loader_constructors.items():
            if loader.validate_extension(filename):
                return ln
        return GenericFITSLoader.name

    def open(self, filename: str, loader: str | None = None, **loader_params):
        loader: str
        if loader is None:
            loader = 'auto'
        if loader not in self.loader_names:
            logger.error(f"Unknown loader type: `{loader}`. Available loaders: {', '.join(self.loader_names)}")
        if loader == 'auto':
            loader = self._get_loader(filename)

        loader: BaseLoader = self._loader_constructors[loader]()
        try:
            loader.open(filename, **loader_params)
        except Exception as e:
            logger.error(f"{type(loader).__name__}: {e} (filename: {filename})")
            return None

        self._loaders[filename] = loader
        logger.debug(f"Database connection opened (filename: {filename})")

        return loader

    def reopen(self, filename: str):
        loader = self._loaders.get(filename)
        if not loader:
            logger.error(f"Failed to re-open connection: connection not open (filename: {filename})")
            return

        loader.reopen(filename)

    def open_image(self, filename: str, loader: str, wcs_source: str | None = None, **kwargs):
        self.open(filename, loader=loader, **kwargs)
        if wcs_source:
            self.open(wcs_source)

    def load(self, filename: str, allowed_dtypes=None, silent: bool = False, lazy: bool = False, **kwargs):
        if not self._loaders.get(filename):
            if not self.open(filename, **kwargs):
                return None, None

        loader = self._loaders.get(filename)
        if lazy and loader.last_data is not None:
            return loader.last_data

        try:
            data, meta = loader.load(**kwargs)
        except Exception as e:
            if not silent:
                logger.error(f"{type(loader).__name__}: {e} (filename: {filename})")
            return None, None

        if allowed_dtypes and not self._validate_dtype(data, allowed_dtypes):
            logger.error(f"Invalid input data type: {type(data)} (filename: {filename})")
            return None, None

        logger.debug(f"Data loaded (filename: {filename})")
        return data, meta

    def close(self, filename: str):
        if not self._loaders.get(filename):
            return
        self._loaders.pop(filename).close()
        logger.debug(f"Database connection closed (filename: {filename})")

    def close_all(self):
        for filename in list(self._loaders):
            self.close(filename)

    @staticmethod
    def _validate_dtype(data, allowed_dtypes: tuple[type, ...]) -> bool:
        if not any(isinstance(data, t) for t in allowed_dtypes):
            return False
        return True


class DataPath(abc.ABC):
    def __init__(self, path: str):
        self._path: str = path

    def __str__(self):
        return self._path

    @property
    def name(self) -> str:
        return self._path

    def resolve(self, obj_id: str | int, cat_entry: Catalog | None):
        field_values = dict(id=obj_id)
        field_names = [fn for _, fn, _, _ in Formatter().parse(self._path) if fn is not None]

        if "id" in field_names:
            field_names.remove("id")

        if field_names:
            if not cat_entry:
                raise ValueError("Catalog entry not loaded")

            for fn in field_names:
                fv = cat_entry.get_col(fn)
                field_values[fn] = fv

        self._path = self._path.format(**field_values)

    def validate(self):
        pass


class LocalPath(DataPath):
    @property
    def name(self) -> str:
        return self._path_obj.name

    @property
    def _path_obj(self) -> pathlib.Path:
        return pathlib.Path(self._path)

    def resolve(self, obj_id: str | int, cat_entry: Catalog | None):
        super().resolve(obj_id, cat_entry)
        self._path = str(self._path_obj.resolve())

    def validate(self):
        if not self._path_obj.exists():
            raise FileNotFoundError


class URLPath(DataPath):
    pass


def add_unit_aliases(unit_aliases: dict[str, list[str]]):
    for unit, aliases in unit_aliases.items():
        u.add_enabled_aliases({alias: u.Unit(unit) for alias in aliases})


def data_browser(default_path, **kwargs) -> FileBrowser:
    return FileBrowser(mode=FileBrowser.OpenDirectory, default_path=default_path, **kwargs)
